# Generated by Django 2.2.7 on 2019-11-08 16:06

from django.db import migrations, models


def move_to_duration(apps, schema_editor):
    # We can't import the model directly as it may be a newer
    # version than this migration expects. We use the historical version.
    duration_models = ['Playbook', 'Play', 'Task', 'Result']
    for duration_model in duration_models:
        model = apps.get_model('api', duration_model)
        for obj in model.objects.all():
            if obj.duration is not None:
                continue
            if obj.ended is not None:
                obj.duration = obj.ended - obj.started
            else:
                obj.duration = obj.updated - obj.started
            obj.save()


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0003_add_missing_result_properties'),
    ]

    operations = [
        migrations.AddField(
            model_name='play',
            name='duration',
            field=models.DurationField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='playbook',
            name='duration',
            field=models.DurationField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='result',
            name='duration',
            field=models.DurationField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='task',
            name='duration',
            field=models.DurationField(blank=True, null=True),
        ),
        migrations.RunPython(move_to_duration)
    ]



#  Copyright (c) 2018 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from rest_framework.test import APITestCase

from ara.api import models, serializers
from ara.api.tests import factories, utils


class RecordTestCase(APITestCase):
    def test_record_factory(self):
        record = factories.RecordFactory(key="test")
        self.assertEqual(record.key, "test")

    def test_record_serializer(self):
        playbook = factories.PlaybookFactory()
        serializer = serializers.RecordSerializer(
            data={"key": "test", "value": factories.RECORD_LIST, "type": "list", "playbook": playbook.id}
        )
        serializer.is_valid()
        record = serializer.save()
        record.refresh_from_db()
        self.assertEqual(record.key, "test")
        self.assertEqual(record.value, utils.compressed_obj(factories.RECORD_LIST))
        self.assertEqual(record.type, "list")

    def test_get_no_records(self):
        request = self.client.get("/api/v1/records")
        self.assertEqual(0, len(request.data["results"]))

    def test_get_record(self):
        record = factories.RecordFactory()
        request = self.client.get("/api/v1/records")
        self.assertEqual(1, len(request.data["results"]))
        self.assertEqual(record.key, request.data["results"][0]["key"])

    def test_delete_record(self):
        record = factories.RecordFactory()
        self.assertEqual(1, models.Record.objects.all().count())
        request = self.client.delete("/api/v1/records/%s" % record.id)
        self.assertEqual(204, request.status_code)
        self.assertEqual(0, models.Record.objects.all().count())

    def test_create_text_record(self):
        playbook = factories.PlaybookFactory()
        self.assertEqual(0, models.Record.objects.count())
        request = self.client.post(
            "/api/v1/records", {"key": "test", "value": "value", "type": "text", "playbook": playbook.id}
        )
        self.assertEqual(201, request.status_code)
        self.assertEqual(1, models.Record.objects.count())

    def test_create_list_record(self):
        playbook = factories.PlaybookFactory()
        self.assertEqual(0, models.Record.objects.count())
        test_list = factories.RECORD_LIST
        request = self.client.post(
            "/api/v1/records", {"key": "listrecord", "value": test_list, "type": "list", "playbook": playbook.id}
        )
        self.assertEqual(201, request.status_code)
        self.assertEqual(1, models.Record.objects.count())
        self.assertEqual(test_list, request.data["value"])

    def test_create_dict_record(self):
        playbook = factories.PlaybookFactory()
        self.assertEqual(0, models.Record.objects.count())
        test_dict = {"a": "dictionary"}
        request = self.client.post(
            "/api/v1/records", {"key": "dictrecord", "value": test_dict, "type": "dict", "playbook": playbook.id}
        )
        self.assertEqual(201, request.status_code)
        self.assertEqual(1, models.Record.objects.count())
        self.assertEqual(test_dict, request.data["value"])

    def test_create_json_record(self):
        playbook = factories.PlaybookFactory()
        self.assertEqual(0, models.Record.objects.count())
        test_json = '{"a": "dictionary"}'
        request = self.client.post(
            "/api/v1/records", {"key": "dictrecord", "value": test_json, "type": "json", "playbook": playbook.id}
        )
        self.assertEqual(201, request.status_code)
        self.assertEqual(1, models.Record.objects.count())
        self.assertEqual(test_json, request.data["value"])

    def test_create_url_record(self):
        playbook = factories.PlaybookFactory()
        self.assertEqual(0, models.Record.objects.count())
        test_url = "https://ara.recordsansible.org"
        request = self.client.post(
            "/api/v1/records", {"key": "dictrecord", "value": test_url, "type": "url", "playbook": playbook.id}
        )
        self.assertEqual(201, request.status_code)
        self.assertEqual(1, models.Record.objects.count())
        self.assertEqual(test_url, request.data["value"])

    def test_partial_update_record(self):
        record = factories.RecordFactory()
        self.assertNotEqual("update", record.key)
        request = self.client.patch("/api/v1/records/%s" % record.id, {"key": "update"})
        self.assertEqual(200, request.status_code)
        record_updated = models.Record.objects.get(id=record.id)
        self.assertEqual("update", record_updated.key)

    def test_get_records_by_playbook(self):
        playbook = factories.PlaybookFactory()
        record = factories.RecordFactory(playbook=playbook, key="by_playbook")
        factories.RecordFactory(key="another_record")
        request = self.client.get("/api/v1/records?playbook=%s" % playbook.id)
        self.assertEqual(2, models.Record.objects.all().count())
        self.assertEqual(1, len(request.data["results"]))
        self.assertEqual(record.key, request.data["results"][0]["key"])
        self.assertEqual(record.playbook.id, request.data["results"][0]["playbook"])

    def test_get_records_by_key(self):
        playbook = factories.PlaybookFactory()
        record = factories.RecordFactory(playbook=playbook, key="by_key")
        factories.RecordFactory(key="another_record")
        request = self.client.get("/api/v1/records?key=%s" % record.key)
        self.assertEqual(2, models.Record.objects.all().count())
        self.assertEqual(1, len(request.data["results"]))
        self.assertEqual(record.key, request.data["results"][0]["key"])
        self.assertEqual(record.playbook.id, request.data["results"][0]["playbook"])

    def test_get_records_by_playbook_and_key(self):
        playbook = factories.PlaybookFactory()
        record = factories.RecordFactory(playbook=playbook, key="by_playbook_and_key")
        factories.RecordFactory(playbook=playbook, key="another_record_in_playbook")
        factories.RecordFactory(key="another_record_in_another_playbook")
        request = self.client.get("/api/v1/records?playbook=%s&key=%s" % (playbook.id, record.key))
        self.assertEqual(3, models.Record.objects.all().count())
        self.assertEqual(1, len(request.data["results"]))
        self.assertEqual(record.key, request.data["results"][0]["key"])
        self.assertEqual(record.playbook.id, request.data["results"][0]["playbook"])

    def test_get_record_by_date(self):
        record = factories.RecordFactory()

        past = datetime.datetime.now() - datetime.timedelta(hours=12)
        negative_date_fields = ["created_before", "updated_before"]
        positive_date_fields = ["created_after", "updated_after"]

        # Expect no record when searching before it was created
        for field in negative_date_fields:
            request = self.client.get("/api/v1/records?%s=%s" % (field, past.isoformat()))
            self.assertEqual(request.data["count"], 0)

        # Expect a record when searching after it was created
        for field in positive_date_fields:
            request = self.client.get("/api/v1/records?%s=%s" % (field, past.isoformat()))
            self.assertEqual(request.data["count"], 1)
            self.assertEqual(request.data["results"][0]["id"], record.id)

    def test_get_record_order(self):
        first_record = factories.RecordFactory(key="alpha")
        second_record = factories.RecordFactory(key="beta")

        # Ensure we have two objects
        request = self.client.get("/api/v1/records")
        self.assertEqual(2, len(request.data["results"]))

        order_fields = ["id", "created", "updated", "key"]
        # Ascending order
        for field in order_fields:
            request = self.client.get("/api/v1/records?order=%s" % field)
            self.assertEqual(request.data["results"][0]["id"], first_record.id)

        # Descending order
        for field in order_fields:
            request = self.client.get("/api/v1/records?order=-%s" % field)
            self.assertEqual(request.data["results"][0]["id"], second_record.id)

    def test_get_playbook_arguments(self):
        record = factories.RecordFactory()
        request = self.client.get("/api/v1/records/%s" % record.id)
        self.assertIn("inventory", request.data["playbook"]["arguments"])



#  Copyright (c) 2019 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

from . import plugins

if __name__ == "__main__":
    print(plugins)



#  Copyright (c) 2019 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict

from rest_framework.pagination import LimitOffsetPagination
from rest_framework.response import Response
from rest_framework.utils.urls import remove_query_param, replace_query_param


class LimitOffsetPaginationWithLinks(LimitOffsetPagination):
    """
    Extends LimitOffsetPagination to provide links
    to first and last pages as well as the limit and offset, if available.
    Generates relative links instead of absolute URIs.
    """

    def get_next_link(self):
        if self.offset + self.limit >= self.count:
            return None

        url = self.request.get_full_path()
        url = replace_query_param(url, self.limit_query_param, self.limit)

        offset = self.offset + self.limit
        return replace_query_param(url, self.offset_query_param, offset)

    def get_previous_link(self):
        if self.offset <= 0:
            return None

        url = self.request.get_full_path()
        url = replace_query_param(url, self.limit_query_param, self.limit)

        if self.offset - self.limit <= 0:
            return remove_query_param(url, self.offset_query_param)

        offset = self.offset - self.limit
        return replace_query_param(url, self.offset_query_param, offset)

    def get_first_link(self):
        if self.offset <= 0:
            return None
        url = self.request.get_full_path()
        return remove_query_param(url, self.offset_query_param)

    def get_last_link(self):
        if self.offset + self.limit >= self.count:
            return None
        url = self.request.get_full_path()
        url = replace_query_param(url, self.limit_query_param, self.limit)
        offset = self.count - self.limit
        return replace_query_param(url, self.offset_query_param, offset)

    def get_paginated_response(self, data):
        return Response(
            OrderedDict(
                [
                    ("count", self.count),
                    ("next", self.get_next_link()),
                    ("previous", self.get_previous_link()),
                    ("first", self.get_first_link()),
                    ("last", self.get_last_link()),
                    ("limit", self.limit),
                    ("offset", self.offset),
                    ("results", data),
                ]
            )
        )



#  Copyright (c) 2018 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from django.utils import timezone
from django.utils.dateparse import parse_duration
from rest_framework.test import APITestCase

from ara.api import models, serializers
from ara.api.tests import factories, utils


class TaskTestCase(APITestCase):
    def test_task_factory(self):
        task = factories.TaskFactory(name="factory")
        self.assertEqual(task.name, "factory")

    def test_task_serializer(self):
        play = factories.PlayFactory()
        file = factories.FileFactory()
        serializer = serializers.TaskSerializer(
            data={
                "name": "serializer",
                "action": "test",
                "lineno": 2,
                "status": "completed",
                "handler": False,
                "play": play.id,
                "file": file.id,
                "playbook": play.playbook.id,
            }
        )
        serializer.is_valid()
        task = serializer.save()
        task.refresh_from_db()
        self.assertEqual(task.name, "serializer")
        self.assertEqual(task.status, "completed")

    def test_task_serializer_compress_tags(self):
        play = factories.PlayFactory()
        file = factories.FileFactory()
        serializer = serializers.TaskSerializer(
            data={
                "name": "compress",
                "action": "test",
                "lineno": 2,
                "status": "running",
                "handler": False,
                "play": play.id,
                "file": file.id,
                "tags": factories.TASK_TAGS,
                "playbook": play.playbook.id,
            }
        )
        serializer.is_valid()
        task = serializer.save()
        task.refresh_from_db()
        self.assertEqual(task.tags, utils.compressed_obj(factories.TASK_TAGS))

    def test_task_serializer_decompress_tags(self):
        task = factories.TaskFactory(tags=utils.compressed_obj(factories.TASK_TAGS))
        serializer = serializers.TaskSerializer(instance=task)
        self.assertEqual(serializer.data["tags"], factories.TASK_TAGS)

    def test_get_no_tasks(self):
        request = self.client.get("/api/v1/tasks")
        self.assertEqual(0, len(request.data["results"]))

    def test_get_tasks(self):
        task = factories.TaskFactory()
        request = self.client.get("/api/v1/tasks")
        self.assertEqual(1, len(request.data["results"]))
        self.assertEqual(task.name, request.data["results"][0]["name"])

    def test_delete_task(self):
        task = factories.TaskFactory()
        self.assertEqual(1, models.Task.objects.all().count())
        request = self.client.delete("/api/v1/tasks/%s" % task.id)
        self.assertEqual(204, request.status_code)
        self.assertEqual(0, models.Task.objects.all().count())

    def test_create_task(self):
        play = factories.PlayFactory()
        file = factories.FileFactory()
        self.assertEqual(0, models.Task.objects.count())
        request = self.client.post(
            "/api/v1/tasks",
            {
                "name": "create",
                "action": "test",
                "lineno": 2,
                "handler": False,
                "status": "running",
                "play": play.id,
                "file": file.id,
                "playbook": play.playbook.id,
            },
        )
        self.assertEqual(201, request.status_code)
        self.assertEqual(1, models.Task.objects.count())

    def test_partial_update_task(self):
        task = factories.TaskFactory()
        self.assertNotEqual("update", task.name)
        request = self.client.patch("/api/v1/tasks/%s" % task.id, {"name": "update"})
        self.assertEqual(200, request.status_code)
        task_updated = models.Task.objects.get(id=task.id)
        self.assertEqual("update", task_updated.name)

    def test_expired_task(self):
        task = factories.TaskFactory(status="running")
        self.assertEqual("running", task.status)

        request = self.client.patch("/api/v1/tasks/%s" % task.id, {"status": "expired"})
        self.assertEqual(200, request.status_code)
        task_updated = models.Task.objects.get(id=task.id)
        self.assertEqual("expired", task_updated.status)

    def test_get_task(self):
        task = factories.TaskFactory()
        request = self.client.get("/api/v1/tasks/%s" % task.id)
        self.assertEqual(task.name, request.data["name"])

    def test_get_tasks_by_playbook(self):
        playbook = factories.PlaybookFactory()
        task = factories.TaskFactory(name="task1", playbook=playbook)
        factories.TaskFactory(name="task2", playbook=playbook)
        request = self.client.get("/api/v1/tasks?playbook=%s" % playbook.id)
        self.assertEqual(2, len(request.data["results"]))
        self.assertEqual(task.name, request.data["results"][1]["name"])
        self.assertEqual("task2", request.data["results"][0]["name"])

    def test_get_tasks_by_name(self):
        # Create a playbook and two tasks
        playbook = factories.PlaybookFactory()
        task = factories.TaskFactory(name="task1", playbook=playbook)
        factories.TaskFactory(name="task2", playbook=playbook)

        # Query for the first task name and expect one result
        request = self.client.get("/api/v1/tasks?name=%s" % task.name)
        self.assertEqual(1, len(request.data["results"]))
        self.assertEqual(task.name, request.data["results"][0]["name"])

    def test_get_task_duration(self):
        started = timezone.now()
        ended = started + datetime.timedelta(hours=1)
        task = factories.TaskFactory(started=started, ended=ended)
        request = self.client.get("/api/v1/tasks/%s" % task.id)
        self.assertEqual(parse_duration(request.data["duration"]), ended - started)

    def test_get_task_by_date(self):
        task = factories.TaskFactory()

        past = datetime.datetime.now() - datetime.timedelta(hours=12)
        negative_date_fields = ["created_before", "started_before", "updated_before"]
        positive_date_fields = ["created_after", "started_after", "updated_after"]

        # Expect no task when searching before it was created
        for field in negative_date_fields:
            request = self.client.get("/api/v1/tasks?%s=%s" % (field, past.isoformat()))
            self.assertEqual(request.data["count"], 0)

        # Expect a task when searching after it was created
        for field in positive_date_fields:
            request = self.client.get("/api/v1/tasks?%s=%s" % (field, past.isoformat()))
            self.assertEqual(request.data["count"], 1)
            self.assertEqual(request.data["results"][0]["id"], task.id)

    def test_get_task_order(self):
        old_started = timezone.now() - datetime.timedelta(hours=12)
        old_ended = old_started + datetime.timedelta(minutes=30)
        old_task = factories.TaskFactory(started=old_started, ended=old_ended)
        new_started = timezone.now() - datetime.timedelta(hours=6)
        new_ended = new_started + datetime.timedelta(hours=1)
        new_task = factories.TaskFactory(started=new_started, ended=new_ended)

        # Ensure we have two objects
        request = self.client.get("/api/v1/tasks")
        self.assertEqual(2, len(request.data["results"]))

        order_fields = ["id", "created", "updated", "started", "ended", "duration"]
        # Ascending order
        for field in order_fields:
            request = self.client.get("/api/v1/tasks?order=%s" % field)
            self.assertEqual(request.data["results"][0]["id"], old_task.id)

        # Descending order
        for field in order_fields:
            request = self.client.get("/api/v1/tasks?order=-%s" % field)
            self.assertEqual(request.data["results"][0]["id"], new_task.id)

    def test_update_wrong_task_status(self):
        task = factories.TaskFactory()
        self.assertNotEqual("wrong", task.status)
        request = self.client.patch("/api/v1/tasks/%s" % task.id, {"status": "wrong"})
        self.assertEqual(400, request.status_code)
        task_updated = models.Task.objects.get(id=task.id)
        self.assertNotEqual("wrong", task_updated.status)

    def test_get_task_by_status(self):
        task = factories.TaskFactory(status="running")
        factories.TaskFactory(status="completed")
        factories.TaskFactory(status="unknown")

        # Confirm we have three objects
        request = self.client.get("/api/v1/tasks")
        self.assertEqual(3, len(request.data["results"]))

        # Test single status
        request = self.client.get("/api/v1/tasks?status=running")
        self.assertEqual(1, len(request.data["results"]))
        self.assertEqual(task.status, request.data["results"][0]["status"])

        # Test multiple status
        request = self.client.get("/api/v1/tasks?status=running&status=completed")
        self.assertEqual(2, len(request.data["results"]))

    def test_get_task_by_action(self):
        task = factories.TaskFactory(action="debug")
        factories.TaskFactory(action="setup")

        # Confirm we have two objects
        request = self.client.get("/api/v1/tasks")
        self.assertEqual(2, len(request.data["results"]))

        # Expect the correct single result when searching
        request = self.client.get("/api/v1/tasks?action=debug")
        self.assertEqual(1, len(request.data["results"]))
        self.assertEqual(task.id, request.data["results"][0]["id"])
        self.assertEqual(task.action, request.data["results"][0]["action"])

    def test_get_task_by_path(self):
        # Create two files with different paths
        first_file = factories.FileFactory(path="/root/roles/foo/tasks/main.yml")
        second_file = factories.FileFactory(path="/root/roles/bar/tasks/main.yml")

        # Create two tasks using these files
        task = factories.TaskFactory(file=first_file)
        factories.TaskFactory(file=second_file)

        # Test exact match
        request = self.client.get("/api/v1/tasks?path=/root/roles/foo/tasks/main.yml")
        self.assertEqual(1, len(request.data["results"]))
        self.assertEqual(task.file.path, request.data["results"][0]["path"])

        # Test partial match
        request = self.client.get("/api/v1/tasks?path=main.yml")
        self.assertEqual(len(request.data["results"]), 2)

    def test_get_playbook_arguments(self):
        task = factories.TaskFactory()
        request = self.client.get("/api/v1/tasks/%s" % task.id)
        self.assertIn("inventory", request.data["playbook"]["arguments"])



# Generated by Django 2.2.1 on 2019-05-17 10:13

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='File',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('path', models.CharField(max_length=255)),
            ],
            options={
                'db_table': 'files',
            },
        ),
        migrations.CreateModel(
            name='FileContent',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('sha1', models.CharField(max_length=40, unique=True)),
                ('contents', models.BinaryField(max_length=4294967295)),
            ],
            options={
                'db_table': 'file_contents',
            },
        ),
        migrations.CreateModel(
            name='Host',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('facts', models.BinaryField(max_length=4294967295)),
                ('alias', models.CharField(max_length=255, null=True)),
                ('changed', models.IntegerField(default=0)),
                ('failed', models.IntegerField(default=0)),
                ('ok', models.IntegerField(default=0)),
                ('skipped', models.IntegerField(default=0)),
                ('unreachable', models.IntegerField(default=0)),
            ],
            options={
                'db_table': 'hosts',
            },
        ),
        migrations.CreateModel(
            name='Label',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'db_table': 'labels',
            },
        ),
        migrations.CreateModel(
            name='Play',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('started', models.DateTimeField(default=django.utils.timezone.now)),
                ('ended', models.DateTimeField(blank=True, null=True)),
                ('name', models.CharField(blank=True, max_length=255, null=True)),
                ('uuid', models.UUIDField()),
                ('status', models.CharField(choices=[('unknown', 'unknown'), ('running', 'running'), ('completed', 'completed')], default='unknown', max_length=25)),
            ],
            options={
                'db_table': 'plays',
            },
        ),
        migrations.CreateModel(
            name='Playbook',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('started', models.DateTimeField(default=django.utils.timezone.now)),
                ('ended', models.DateTimeField(blank=True, null=True)),
                ('name', models.CharField(max_length=255, null=True)),
                ('ansible_version', models.CharField(max_length=255)),
                ('status', models.CharField(choices=[('unknown', 'unknown'), ('running', 'running'), ('completed', 'completed'), ('failed', 'failed')], default='unknown', max_length=25)),
                ('arguments', models.BinaryField(max_length=4294967295)),
                ('path', models.CharField(max_length=255)),
                ('labels', models.ManyToManyField(to='api.Label')),
            ],
            options={
                'db_table': 'playbooks',
            },
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('started', models.DateTimeField(default=django.utils.timezone.now)),
                ('ended', models.DateTimeField(blank=True, null=True)),
                ('name', models.TextField(blank=True, null=True)),
                ('action', models.TextField()),
                ('lineno', models.IntegerField()),
                ('tags', models.BinaryField(max_length=4294967295)),
                ('handler', models.BooleanField()),
                ('status', models.CharField(choices=[('unknown', 'unknown'), ('running', 'running'), ('completed', 'completed')], default='unknown', max_length=25)),
                ('file', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tasks', to='api.File')),
                ('play', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tasks', to='api.Play')),
                ('playbook', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tasks', to='api.Playbook')),
            ],
            options={
                'db_table': 'tasks',
            },
        ),
        migrations.CreateModel(
            name='Result',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('started', models.DateTimeField(default=django.utils.timezone.now)),
                ('ended', models.DateTimeField(blank=True, null=True)),
                ('status', models.CharField(choices=[('ok', 'ok'), ('failed', 'failed'), ('skipped', 'skipped'), ('unreachable', 'unreachable'), ('changed', 'changed'), ('ignored', 'ignored'), ('unknown', 'unknown')], default='unknown', max_length=25)),
                ('content', models.BinaryField(max_length=4294967295)),
                ('host', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='api.Host')),
                ('play', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='api.Play')),
                ('playbook', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='api.Playbook')),
                ('task', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='api.Task')),
            ],
            options={
                'db_table': 'results',
            },
        ),
        migrations.AddField(
            model_name='play',
            name='playbook',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='plays', to='api.Playbook'),
        ),
        migrations.AddField(
            model_name='host',
            name='playbook',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='hosts', to='api.Playbook'),
        ),
        migrations.AddField(
            model_name='file',
            name='content',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='files', to='api.FileContent'),
        ),
        migrations.AddField(
            model_name='file',
            name='playbook',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='files', to='api.Playbook'),
        ),
        migrations.CreateModel(
            name='Record',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('key', models.CharField(max_length=255)),
                ('value', models.BinaryField(max_length=4294967295)),
                ('type', models.CharField(max_length=255)),
                ('playbook', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='records', to='api.Playbook')),
            ],
            options={
                'db_table': 'records',
                'unique_together': {('key', 'playbook')},
            },
        ),
        migrations.AlterUniqueTogether(
            name='host',
            unique_together={('name', 'playbook')},
        ),
        migrations.AlterUniqueTogether(
            name='file',
            unique_together={('path', 'playbook')},
        ),
    ]



#  Copyright (c) 2019 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA Records Ansible is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA Records Ansible is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA Records Ansible. If not, see <http://www.gnu.org/licenses/>.

import collections
import hashlib
import json
import zlib

from rest_framework import serializers

from ara.api import models

# Constants used for defaults which rely on compression so we don't need to
# reproduce this code elsewhere.
EMPTY_DICT = zlib.compress(json.dumps({}).encode("utf8"))
EMPTY_LIST = zlib.compress(json.dumps([]).encode("utf8"))
EMPTY_STRING = zlib.compress(json.dumps("").encode("utf8"))


class CompressedTextField(serializers.CharField):
    """
    Compresses text before storing it in the database.
    Decompresses text from the database before serving it.
    """

    def to_representation(self, obj):
        return zlib.decompress(obj).decode("utf8")

    def to_internal_value(self, data):
        return zlib.compress(data.encode("utf8"))


class CompressedObjectField(serializers.JSONField):
    """
    Serializes/compresses an object (i.e, list, dict) before storing it in the
    database.
    Decompresses/deserializes an object before serving it.
    """

    def to_representation(self, obj):
        return json.loads(zlib.decompress(obj).decode("utf8"))

    def to_internal_value(self, data):
        return zlib.compress(json.dumps(data).encode("utf8"))


class FileContentField(serializers.CharField):
    """
    Compresses text before storing it in the database.
    Decompresses text from the database before serving it.
    """

    def to_representation(self, obj):
        return zlib.decompress(obj.contents).decode("utf8")

    def to_internal_value(self, data):
        contents = data.encode("utf8")
        sha1 = hashlib.sha1(contents).hexdigest()
        content_file, created = models.FileContent.objects.get_or_create(
            sha1=sha1, defaults={"sha1": sha1, "contents": zlib.compress(contents)}
        )
        return content_file


class CreatableSlugRelatedField(serializers.SlugRelatedField):
    """
    A SlugRelatedField that supports get_or_create.
    Used for creating or retrieving labels by name.
    """

    def to_representation(self, obj):
        return {"id": obj.id, "name": obj.name}

    # Overriding RelatedField.to_representation causes error in Browseable API
    # https://github.com/encode/django-rest-framework/issues/5141
    def get_choices(self, cutoff=None):
        queryset = self.get_queryset()
        if queryset is None:
            # Ensure that field.choices returns something sensible
            # even when accessed with a read-only field.
            return {}

        if cutoff is not None:
            queryset = queryset[:cutoff]

        return collections.OrderedDict(
            [
                (
                    # This is the only line that differs
                    # from the RelatedField's implementation
                    item.pk,
                    self.display_value(item),
                )
                for item in queryset
            ]
        )

    def to_internal_value(self, data):
        try:
            return self.get_queryset().get_or_create(**{self.slug_field: data})[0]
        except (TypeError, ValueError):
            self.fail("invalid")



#  Copyright (c) 2018 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from django.db.utils import IntegrityError
from rest_framework.test import APITestCase

from ara.api import models, serializers
from ara.api.tests import factories


class LabelTestCase(APITestCase):
    def test_label_factory(self):
        label = factories.LabelFactory(name="factory")
        self.assertEqual(label.name, "factory")

    def test_label_serializer(self):
        serializer = serializers.LabelSerializer(data={"name": "serializer"})
        serializer.is_valid()
        label = serializer.save()
        label.refresh_from_db()
        self.assertEqual(label.name, "serializer")

    def test_create_label(self):
        self.assertEqual(0, models.Label.objects.count())
        request = self.client.post("/api/v1/labels", {"name": "compress"})
        self.assertEqual(201, request.status_code)
        self.assertEqual(1, models.Label.objects.count())

    def test_get_no_labels(self):
        request = self.client.get("/api/v1/labels")
        self.assertEqual(0, len(request.data["results"]))

    def test_get_labels(self):
        label = factories.LabelFactory()
        request = self.client.get("/api/v1/labels")
        self.assertEqual(1, len(request.data["results"]))
        self.assertEqual(label.name, request.data["results"][0]["name"])

    def test_get_label(self):
        label = factories.LabelFactory()
        request = self.client.get("/api/v1/labels/%s" % label.id)
        self.assertEqual(label.name, request.data["name"])

    def test_partial_update_label(self):
        label = factories.LabelFactory()
        self.assertNotEqual("updated", label.name)
        request = self.client.patch("/api/v1/labels/%s" % label.id, {"name": "updated"})
        self.assertEqual(200, request.status_code)
        label_updated = models.Label.objects.get(id=label.id)
        self.assertEqual("updated", label_updated.name)

    def test_delete_label(self):
        label = factories.LabelFactory()
        self.assertEqual(1, models.Label.objects.all().count())
        request = self.client.delete("/api/v1/labels/%s" % label.id)
        self.assertEqual(204, request.status_code)
        self.assertEqual(0, models.Label.objects.all().count())

    def test_get_label_by_date(self):
        label = factories.LabelFactory()

        past = datetime.datetime.now() - datetime.timedelta(hours=12)
        negative_date_fields = ["created_before", "updated_before"]
        positive_date_fields = ["created_after", "updated_after"]

        # Expect no label when searching before it was created
        for field in negative_date_fields:
            request = self.client.get("/api/v1/labels?%s=%s" % (field, past.isoformat()))
            self.assertEqual(request.data["count"], 0)

        # Expect a label when searching after it was created
        for field in positive_date_fields:
            request = self.client.get("/api/v1/labels?%s=%s" % (field, past.isoformat()))
            self.assertEqual(request.data["count"], 1)
            self.assertEqual(request.data["results"][0]["id"], label.id)

    def test_get_label_order(self):
        first_label = factories.LabelFactory(name="first")
        second_label = factories.LabelFactory(name="second")

        # Ensure we have two objects
        request = self.client.get("/api/v1/labels")
        self.assertEqual(2, len(request.data["results"]))

        order_fields = ["id", "created", "updated"]
        # Ascending order
        for field in order_fields:
            request = self.client.get("/api/v1/labels?order=%s" % field)
            self.assertEqual(request.data["results"][0]["name"], first_label.name)

        # Descending order
        for field in order_fields:
            request = self.client.get("/api/v1/labels?order=-%s" % field)
            self.assertEqual(request.data["results"][0]["name"], second_label.name)

    def test_unique_label_names(self):
        # Create a first label
        factories.LabelFactory(name="label")
        with self.assertRaises(IntegrityError):
            # Creating a second label with the same name should yield an exception
            factories.LabelFactory(name="label")



import codecs
import os
import shutil

from django.core.management.base import BaseCommand
from django.template.loader import render_to_string

from ara.api import models, serializers


class Command(BaseCommand):
    help = "Generates a static tree of the web application"
    DEFAULT_PARAMS = dict(static_generation=True)
    rendered = 0

    @staticmethod
    def create_dirs(path):
        # create main output dir
        if not os.path.exists(path):
            os.mkdir(path)

        # create subdirs
        dirs = ["playbooks", "files", "hosts", "results", "records"]
        for dir in dirs:
            if not os.path.exists(os.path.join(path, dir)):
                os.mkdir(os.path.join(path, dir))

        # Retrieve static assets (../../static)
        shutil.rmtree(os.path.join(path, "static"), ignore_errors=True)
        ui_path = os.path.abspath(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
        shutil.copytree(os.path.join(ui_path, "static"), os.path.join(path, "static"))
        # copy robots.txt from templates to root directory
        shutil.copyfile(os.path.join(ui_path, "templates/robots.txt"), os.path.join(path, "robots.txt"))

    def add_arguments(self, parser):
        parser.add_argument("path", help="Path where the static files will be built in", type=str)

    def render(self, template, destination, **kwargs):
        self.rendered += 1
        with open(destination, "w") as f:
            f.write(render_to_string(template, kwargs))

    def handle(self, *args, **options):
        path = options.get("path")
        self.create_dirs(path)

        # TODO: Leverage ui views directly instead of duplicating logic here
        query = models.Playbook.objects.all().order_by("-id")
        serializer = serializers.ListPlaybookSerializer(query, many=True)

        print("[ara] Generating static files for %s playbooks at %s..." % (query.count(), path))

        # Index
        destination = os.path.join(path, "index.html")
        data = {"data": {"results": serializer.data}, "page": "index", **self.DEFAULT_PARAMS}
        self.render("index.html", destination, **data)

        # Escape surrogates to prevent UnicodeEncodeError exceptions
        codecs.register_error("strict", codecs.lookup_error("surrogateescape"))

        # Playbooks
        for pb in query:
            playbook = serializers.DetailedPlaybookSerializer(pb)
            hosts = serializers.ListHostSerializer(
                models.Host.objects.filter(playbook=playbook.data["id"]).order_by("name").all(), many=True
            )
            files = serializers.ListFileSerializer(
                models.File.objects.filter(playbook=playbook.data["id"]).all(), many=True
            )
            records = serializers.ListRecordSerializer(
                models.Record.objects.filter(playbook=playbook.data["id"]).all(), many=True
            )
            results = serializers.ListResultSerializer(
                models.Result.objects.filter(playbook=playbook.data["id"]).all(), many=True
            )

            # Backfill task and host data into results
            for result in results.data:
                task_id = result["task"]
                result["task"] = serializers.SimpleTaskSerializer(models.Task.objects.get(pk=task_id)).data
                host_id = result["host"]
                result["host"] = serializers.SimpleHostSerializer(models.Host.objects.get(pk=host_id)).data

            # Results are paginated in the dynamic version and the template expects data in a specific format
            formatted_results = {"count": len(results.data), "results": results.data}

            destination = os.path.join(path, "playbooks/%s.html" % playbook.data["id"])
            self.render(
                "playbook.html",
                destination,
                playbook=playbook.data,
                hosts=hosts.data,
                files=files.data,
                records=records.data,
                results=formatted_results,
                current_page_results=None,
                search_form=None,
                page="playbook",
                **self.DEFAULT_PARAMS
            )

        # Files
        query = models.File.objects.all()
        for file in query.all():
            destination = os.path.join(path, "files/%s.html" % file.id)
            serializer = serializers.DetailedFileSerializer(file)
            data = {"file": serializer.data, "page": "file", **self.DEFAULT_PARAMS}
            self.render("file.html", destination, **data)

        # Hosts
        query = models.Host.objects.all()
        for host in query.all():
            destination = os.path.join(path, "hosts/%s.html" % host.id)
            serializer = serializers.DetailedHostSerializer(host)

            # fmt: off
            host_results = serializers.ListResultSerializer(
                models.Result.objects.filter(host=host.id).all(), many=True
            )
            # fmt: on

            # Backfill task data into results
            for result in host_results.data:
                task_id = result["task"]
                result["task"] = serializers.SimpleTaskSerializer(models.Task.objects.get(pk=task_id)).data

            # Results are paginated in the dynamic version and the template expects data in a specific format
            formatted_results = {"count": len(host_results.data), "results": host_results.data}

            self.render(
                "host.html",
                destination,
                current_page_results=None,
                host=serializer.data,
                page="host",
                results=formatted_results,
                search_form=None,
                **self.DEFAULT_PARAMS
            )

        # Results
        query = models.Result.objects.all()
        for result in query.all():
            destination = os.path.join(path, "results/%s.html" % result.id)
            serializer = serializers.DetailedResultSerializer(result)
            data = {"result": serializer.data, "page": "result", **self.DEFAULT_PARAMS}
            self.render("result.html", destination, **data)

        # Records
        query = models.Record.objects.all()
        for record in query.all():
            destination = os.path.join(path, "records/%s.html" % record.id)
            serializer = serializers.DetailedRecordSerializer(record)
            data = {"record": serializer.data, "page": "record", **self.DEFAULT_PARAMS}
            self.render("record.html", destination, **data)

        print("[ara] %s files generated." % self.rendered)



# Generated by Django 2.2.1 on 2019-05-23 17:34

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='host',
            name='alias',
        ),
    ]



#  Copyright (c) 2019 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

import threading

from django.db.backends.sqlite3.base import DatabaseWrapper as BaseDatabaseWrapper

local_storage = threading.local()


class DatabaseWrapper(BaseDatabaseWrapper):
    """
    Custom sqlite database backend meant to work with ara.server.wsgi.distributed_sqlite
    in order to dynamically load different databases at runtime.
    """

    def get_new_connection(self, conn_params):
        if hasattr(local_storage, "db_path") and local_storage.db_path:
            conn_params["database"] = local_storage.db_path
        return super().get_new_connection(conn_params)



# Generated by Django 2.2.16 on 2020-09-06 18:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0005_unique_label_names'),
    ]

    # Previously, choices included "ignored" and "changed" but these were never used
    # See: https://github.com/ansible-community/ara/issues/150
    operations = [
        migrations.AlterField(
            model_name='result',
            name='status',
            field=models.CharField(choices=[('ok', 'ok'), ('failed', 'failed'), ('skipped', 'skipped'), ('unreachable', 'unreachable'), ('unknown', 'unknown')], default='unknown', max_length=25),
        ),
    ]



# Generated by Django 2.2.17 on 2020-12-04 04:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0007_add_expired_status'),
    ]

    operations = [
        migrations.AddField(
            model_name='playbook',
            name='controller',
            field=models.CharField(default='localhost', max_length=255),
        ),
    ]



#  Copyright (c) 2019 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

import pkg_resources
from rest_framework.test import APITestCase


class RootTestCase(APITestCase):
    def test_root_endpoint(self):
        result = self.client.get("/api/")
        self.assertEqual(set(result.data.keys()), set(["kind", "version", "api"]))
        self.assertEqual(result.data["kind"], "ara")
        self.assertEqual(result.data["version"], pkg_resources.get_distribution("ara").version)
        self.assertTrue(len(result.data["api"]), 1)
        self.assertTrue(result.data["api"][0].endswith("/api/v1/"))



#  Copyright (c) 2018 Red Hat, Inc.
#
#  This file is part of ARA: Ansible Run Analysis.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

# This is an "offline" API client that does not require standing up
# an API server and does not execute actual HTTP calls.

import logging
import os
import threading

from ara.clients.http import AraHttpClient
from ara.setup.exceptions import MissingDjangoException

try:
    from django.core.handlers.wsgi import WSGIHandler
    from django.core.servers.basehttp import ThreadedWSGIServer, WSGIRequestHandler
except ImportError as e:
    raise MissingDjangoException from e


class AraOfflineClient(AraHttpClient):
    def __init__(self, auth=None, run_sql_migrations=True):
        self.log = logging.getLogger(__name__)

        from django import setup as django_setup
        from django.core.management import execute_from_command_line

        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "ara.server.settings")

        if run_sql_migrations:
            # Automatically create the database and run migrations (is there a better way?)
            execute_from_command_line(["django", "migrate"])

        # Set up the things Django needs
        django_setup()

        self._start_server()
        super().__init__(endpoint="http://localhost:%d" % self.server_thread.port, auth=auth)

    def _start_server(self):
        self.server_thread = ServerThread("localhost")
        self.server_thread.start()

        # Wait for the live server to be ready
        self.server_thread.is_ready.wait()
        if self.server_thread.error:
            raise self.server_thread.error


class ServerThread(threading.Thread):
    def __init__(self, host, port=0):
        self.host = host
        self.port = port
        self.is_ready = threading.Event()
        self.error = None
        super().__init__(daemon=True)

    def run(self):
        """
        Set up the live server and databases, and then loop over handling
        HTTP requests.
        """
        try:
            # Create the handler for serving static and media files
            self.httpd = self._create_server()
            # If binding to port zero, assign the port allocated by the OS.
            if self.port == 0:
                self.port = self.httpd.server_address[1]
            self.httpd.set_app(WSGIHandler())
            self.is_ready.set()
            self.httpd.serve_forever()
        except Exception as e:
            self.error = e
            self.is_ready.set()

    def _create_server(self):
        return ThreadedWSGIServer((self.host, self.port), QuietWSGIRequestHandler, allow_reuse_address=False)


class QuietWSGIRequestHandler(WSGIRequestHandler):
    def log_message(*args):
        pass



#  Copyright (c) 2018 Red Hat, Inc.
#
#  This file is part of ARA: Ansible Run Analysis.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

# This is an "offline" API client that does not require standing up
# an API server and does not execute actual HTTP calls.

import json
import logging
import weakref

import pbr.version
import requests

from ara.clients.utils import active_client

CLIENT_VERSION = pbr.version.VersionInfo("ara").release_string()


class HttpClient(object):
    def __init__(self, endpoint="http://127.0.0.1:8000", auth=None, timeout=30, verify=True):
        self.log = logging.getLogger(__name__)

        self.endpoint = endpoint.rstrip("/")
        self.auth = auth
        self.timeout = int(timeout)
        self.verify = verify
        self.headers = {
            "User-Agent": "ara-http-client_%s" % CLIENT_VERSION,
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        self.http = requests.Session()
        self.http.headers.update(self.headers)
        if self.auth is not None:
            self.http.auth = self.auth
        self.http.verify = self.verify

    def _request(self, method, url, **payload):
        # Use requests.Session to do the query
        # The actual endpoint is:
        # <endpoint>              <url>
        # http://127.0.0.1:8000 / api/v1/playbooks
        return self.http.request(method, self.endpoint + url, timeout=self.timeout, **payload)

    def get(self, url, **payload):
        if payload:
            return self._request("get", url, **payload)
        else:
            return self._request("get", url)

    def patch(self, url, **payload):
        return self._request("patch", url, data=json.dumps(payload))

    def post(self, url, **payload):
        return self._request("post", url, data=json.dumps(payload))

    def put(self, url, **payload):
        return self._request("put", url, data=json.dumps(payload))

    def delete(self, url):
        return self._request("delete", url)


class AraHttpClient(object):
    def __init__(self, endpoint="http://127.0.0.1:8000", auth=None, timeout=30, verify=True):
        self.log = logging.getLogger(__name__)
        self.endpoint = endpoint
        self.auth = auth
        self.timeout = int(timeout)
        self.verify = verify
        self.client = HttpClient(endpoint=self.endpoint, timeout=self.timeout, auth=self.auth, verify=self.verify)
        active_client._instance = weakref.ref(self)

    def _request(self, method, url, **kwargs):
        func = getattr(self.client, method)
        if method == "delete":
            response = func(url)
        else:
            response = func(url, **kwargs)

        if response.status_code >= 500:
            self.log.error("Failed to {method} on {url}: {content}".format(method=method, url=url, content=kwargs))

        self.log.debug("HTTP {status}: {method} on {url}".format(status=response.status_code, method=method, url=url))

        if response.status_code not in [200, 201, 204]:
            self.log.error("Failed to {method} on {url}: {content}".format(method=method, url=url, content=kwargs))

        if response.status_code == 204:
            return response

        return response.json()

    def get(self, endpoint, **kwargs):
        return self._request("get", endpoint, params=kwargs)

    def patch(self, endpoint, **kwargs):
        return self._request("patch", endpoint, **kwargs)

    def post(self, endpoint, **kwargs):
        return self._request("post", endpoint, **kwargs)

    def put(self, endpoint, **kwargs):
        return self._request("put", endpoint, **kwargs)

    def delete(self, endpoint, **kwargs):
        return self._request("delete", endpoint)



#  Copyright (c) 2019 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from django import template
from django.utils.dateparse import parse_datetime

register = template.Library()


@register.filter(name="format_duration")
def format_duration(duration):
    if duration is not None:
        return duration[:-4]
    return duration


@register.filter(name="format_date")
def format_datetime(datetime):
    return parse_datetime(datetime).strftime("%d %b %Y %H:%M:%S %z")


@register.simple_tag(name="past_timestamp")
def past_timestamp(weeks=0, days=0, hours=0, minutes=0, seconds=0):
    """
    Produces a timestamp from the past compatible with the API.
    Used to provide time ranges by templates.
    Expects a dictionary of arguments to timedelta, for example:
        datetime.timedelta(hours=24)
        datetime.timedelta(days=7)
    See: https://docs.python.org/3/library/datetime.html#datetime.timedelta
    """
    delta = dict()
    if weeks:
        delta["weeks"] = weeks
    if days:
        delta["days"] = days
    if hours:
        delta["hours"] = hours
    if minutes:
        delta["minutes"] = minutes
    if seconds:
        delta["seconds"] = seconds

    return (datetime.datetime.now() - datetime.timedelta(**delta)).isoformat()



#  Copyright (c) 2019 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

from django import forms

from ara.api import models


class PlaybookSearchForm(forms.Form):
    ansible_version = forms.CharField(label="Ansible version", max_length=255, required=False)
    controller = forms.CharField(label="Playbook controller", max_length=255, required=False)
    name = forms.CharField(label="Playbook name", max_length=255, required=False)
    path = forms.CharField(label="Playbook path", max_length=255, required=False)
    status = forms.MultipleChoiceField(
        widget=forms.CheckboxSelectMultiple, choices=models.Playbook.STATUS, required=False
    )
    label = forms.CharField(label="Playbook label", max_length=255, required=False)
    started_after = forms.DateField(label="Started after", required=False)
    order = forms.CharField(label="Order", max_length=255, required=False)


class ResultSearchForm(forms.Form):
    host = forms.CharField(label="Host id", max_length=10, required=False)
    task = forms.CharField(label="Task id", max_length=10, required=False)
    changed = forms.BooleanField(label="Changed", required=False)

    status = forms.MultipleChoiceField(
        widget=forms.CheckboxSelectMultiple, choices=models.Result.STATUS, required=False
    )



# Copyright (c) 2020 The ARA Records Ansible authors
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# See https://github.com/ansible-community/ara/issues/26 for rationale on expiring

import logging
import os
import sys
from datetime import datetime, timedelta

from cliff.command import Command

from ara.cli.base import global_arguments
from ara.clients.utils import get_client


class ExpireObjects(Command):
    """ Expires objects that have been in the running state for too long """

    log = logging.getLogger(__name__)
    expired = 0

    def get_parser(self, prog_name):
        parser = super(ExpireObjects, self).get_parser(prog_name)
        parser = global_arguments(parser)
        # fmt: off
        parser.add_argument(
            "--hours",
            type=int,
            default=24,
            help="Expires objects that have been running state for this many hours (default: 24)"
        )
        parser.add_argument(
            "--order",
            metavar="<order>",
            default="started",
            help=(
                "Orders objects by a field ('id', 'created', 'updated', 'started', 'ended')\n"
                "Defaults to 'started' descending so the oldest objects would be expired first.\n"
                "The order can be reversed by using '-': ara expire --order=-started"
            ),
        )
        parser.add_argument(
            "--limit",
            metavar="<limit>",
            default=os.environ.get("ARA_CLI_LIMIT", 200),
            help=("Only expire the first <limit> determined by the ordering. Defaults to ARA_CLI_LIMIT or 200.")
        )
        parser.add_argument(
            "--confirm",
            action="store_true",
            help="Confirm expiration of objects, otherwise runs without expiring any objects",
        )
        # fmt: on
        return parser

    def take_action(self, args):
        client = get_client(
            client=args.client,
            endpoint=args.server,
            timeout=args.timeout,
            username=args.username,
            password=args.password,
            verify=False if args.insecure else True,
            run_sql_migrations=False,
        )

        if not args.confirm:
            self.log.info("--confirm was not specified, no objects will be expired")

        query = dict(status="running")
        # generate a timestamp from n days ago in a format we can query the API with
        # ex: 2019-11-21T00:57:41.702229
        query["updated_before"] = (datetime.now() - timedelta(hours=args.hours)).isoformat()
        query["order"] = args.order
        query["limit"] = args.limit

        endpoints = ["/api/v1/playbooks", "/api/v1/plays", "/api/v1/tasks"]
        for endpoint in endpoints:
            objects = client.get(endpoint, **query)
            self.log.info("Found %s objects matching query on %s" % (objects["count"], endpoint))
            # TODO: Improve client validation and exception handling
            if "count" not in objects:
                # If we didn't get an answer we can parse, it's probably due to an error 500, 403, 401, etc.
                # The client would have logged the error.
                self.log.error(
                    "Client failed to retrieve results, see logs for ara.clients.offline or ara.clients.http."
                )
                sys.exit(1)

            for obj in objects["results"]:
                link = "%s/%s" % (endpoint, obj["id"])
                if not args.confirm:
                    self.log.info(
                        "Dry-run: %s would have been expired, status is running since %s" % (link, obj["updated"])
                    )
                else:
                    self.log.info("Expiring %s, status is running since %s" % (link, obj["updated"]))
                    client.patch(link, status="expired")
                    self.expired += 1

        self.log.info("%s objects expired" % self.expired)



#  Copyright (c) 2019 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os

from ara.setup.exceptions import MissingDjangoException

try:
    from django.core.handlers.wsgi import get_path_info, get_script_name
    from django.core.wsgi import get_wsgi_application
except ImportError as e:
    raise MissingDjangoException from e

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "ara.server.settings")

logger = logging.getLogger(__name__)

# The default WSGI application
application = get_wsgi_application()


def handle_405(start_response):
    start_response("405 Method Not Allowed", [("content-type", "text/html")])
    return [b"<h1>Method Not Allowed</h1><p>This endpoint is read only.</p>"]


def handle_404(start_response):
    start_response("404 Not Found", [("content-type", "text/html")])
    return [b"<h1>Not Found</h1><p>The requested resource was not found on this server.</p>"]


def distributed_sqlite(environ, start_response):
    """
    Custom WSGI application meant to work with ara.server.db.backends.distributed_sqlite
    in order to dynamically load different databases at runtime.
    """
    # This endpoint is read only, do not accept write requests.
    if environ["REQUEST_METHOD"] not in ["GET", "HEAD", "OPTIONS"]:
        handle_405(start_response)

    script_name = get_script_name(environ)
    path_info = get_path_info(environ)

    from django.conf import settings

    # The root under which database files are expected
    root = settings.DISTRIBUTED_SQLITE_ROOT
    # The prefix after which everything should be delegated (ex: /ara-report)
    prefix = settings.DISTRIBUTED_SQLITE_PREFIX

    # Static assets should always be served by the regular app
    if path_info.startswith(settings.STATIC_URL):
        return application(environ, start_response)

    if prefix not in path_info:
        logger.warn("Ignoring request: URL does not contain delegated prefix (%s)" % prefix)
        return handle_404(start_response)

    # Slice path_info up until after the prefix to obtain the requested directory
    i = path_info.find(prefix) + len(prefix)
    fs_path = path_info[:i]

    # Make sure we aren't escaping outside the root and the directory exists
    db_dir = os.path.abspath(os.path.join(root, fs_path.lstrip("/")))
    if not db_dir.startswith(root):
        logger.warn("Ignoring request: path is outside the root (%s)" % db_dir)
        return handle_404(start_response)
    elif not os.path.exists(db_dir):
        logger.warn("Ignoring request: database directory not found (%s)" % db_dir)
        return handle_404(start_response)

    # Find the database file and make sure it exists
    db_file = os.path.join(db_dir, "ansible.sqlite")
    if not os.path.exists(db_file):
        logger.warn("Ignoring request: database file not found (%s)" % db_file)
        return handle_404(start_response)

    # Tell Django about the new URLs it should be using
    environ["SCRIPT_NAME"] = script_name + fs_path
    environ["PATH_INFO"] = path_info[len(fs_path) :]  # noqa: E203

    # Store the path of the database in a thread so the distributed_sqlite
    # database backend can retrieve it.
    from ara.server.db.backends.distributed_sqlite.base import local_storage

    local_storage.db_path = db_file
    try:
        return application(environ, start_response)
    finally:
        del local_storage.db_path



# Copyright (c) 2020 The ARA Records Ansible authors
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import logging
import os
import sys
from datetime import datetime, timedelta

from cliff.command import Command

import ara.cli.utils as cli_utils
from ara.cli.base import global_arguments
from ara.clients.utils import get_client

from prometheus_client import start_http_server, Summary

class PlayDelete(Command):
    """ Deletes the specified play and associated resources """

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(PlayDelete, self).get_parser(prog_name)
        parser = global_arguments(parser)
        # fmt: off
        # TODO: Arguments
        # fmt: on
        return parser

    def take_action(self, args):
        client = get_client(
            client=args.client,
            endpoint=args.server,
            timeout=args.timeout,
            username=args.username,
            password=args.password,
            verify=False if args.insecure else True,
            run_sql_migrations=False,
        )

        # TODO: Improve client to be better at handling exceptions
        # Get metrics
        # playbooks per ansible version
        # playbooks per controller
        # playbook duration
        # number of playbooks
        # number of plays
        # number of tasks
        # number of results
        # number of hosts

        playbooks = client.get("/api/v1/playbooks", limit=1)["count"]
        plays = client.get("/api/v1/plays", limit=1)["count"]
        hosts = client.get("/api/v1/hosts", limit=1)["count"]
        tasks = client.get("/api/v1/tasks", limit=1)["count"]
        results = client.get("/api/v1/results", limit=1)["count"]
        records = client.get("/api/v1/records", limit=1)["count"]

def main():
    parser = argparse.ArgumentParser()
    

client = get_client(
    client=args.client,
    endpoint=args.server,
    timeout=args.timeout,
    username=args.username,
    password=args.password,
    verify=False if args.insecure else True,
    run_sql_migrations=False,
)


# Create a metric to track time spent and requests made.
REQUEST_TIME = Summary('request_processing_seconds', 'Time spent processing request')

# Decorate function with metric.
@REQUEST_TIME.time()
def process_request(t):
    """A dummy function that takes some time."""
    time.sleep(t)

if __name__ == '__main__':
    # Start up the server to expose the metrics.
    start_http_server(8000)


#  Copyright (c) 2020 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

import os

from django import template
from django.template.defaultfilters import stringfilter

register = template.Library()


@register.filter(name="truncatepath")
@stringfilter
def truncatepath(path, count):
    """
    Truncates a path to less than 'count' characters.
    Paths are truncated on path separators.
    We prepend an ellipsis when we return a truncated path.
    """
    try:
        length = int(count)
    except ValueError:
        return path

    # Return immediately if there's nothing to truncate
    if len(path) < length:
        return path

    dirname, basename = os.path.split(path)
    while dirname:
        if len(dirname) + len(basename) < length:
            break
        dirlist = dirname.split("/")
        dirlist.pop(0)
        dirname = "/".join(dirlist)

    return "..." + os.path.join(dirname, basename)



# Copyright (c) 2020 The ARA Records Ansible authors
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import logging
import os
import sys

from cliff.command import Command
from cliff.lister import Lister
from cliff.show import ShowOne

import ara.cli.utils as cli_utils
from ara.cli.base import global_arguments
from ara.clients.utils import get_client


class HostList(Lister):
    """ Returns a list of hosts based on search queries """

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(HostList, self).get_parser(prog_name)
        parser = global_arguments(parser)
        # fmt: off
        # Host search arguments and ordering as per ara.api.filters.HostFilter
        # TODO: non-exhaustive (searching for failed, ok, unreachable, etc.)
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=("List hosts matching the provided name (full or partial)"),
        )
        parser.add_argument(
            "--playbook",
            metavar="<playbook_id>",
            default=None,
            help=("List hosts for a specified playbook id"),
        )

        changed = parser.add_mutually_exclusive_group()
        changed.add_argument(
            "--with-changed",
            action="store_true",
            default=False,
            help=("Return hosts with changed results")
        )
        changed.add_argument(
            "--without-changed",
            action="store_true",
            default=False,
            help=("Don't return hosts with changed results")
        )

        failed = parser.add_mutually_exclusive_group()
        failed.add_argument(
            "--with-failed",
            action="store_true",
            default=False,
            help=("Return hosts with failed results")
        )
        failed.add_argument(
            "--without-failed",
            action="store_true",
            default=False,
            help=("Don't return hosts with failed results")
        )

        unreachable = parser.add_mutually_exclusive_group()
        unreachable.add_argument(
            "--with-unreachable",
            action="store_true",
            default=False,
            help=("Return hosts with unreachable results")
        )
        unreachable.add_argument(
            "--without-unreachable",
            action="store_true",
            default=False,
            help=("Don't return hosts with unreachable results")
        )
        parser.add_argument(
            "--resolve",
            action="store_true",
            default=os.environ.get("ARA_CLI_RESOLVE", False),
            help=("Resolve IDs to identifiers (such as path or names). Defaults to ARA_CLI_RESOLVE or False")
        )
        parser.add_argument(
            "--long",
            action="store_true",
            default=False,
            help=("Don't truncate paths")
        )
        parser.add_argument(
            "--order",
            metavar="<order>",
            default="-updated",
            help=(
                "Orders hosts by a field ('id', 'created', 'updated', 'name')\n"
                "Defaults to '-updated' descending so the most recent host is at the top.\n"
                "The order can be reversed by omitting the '-': ara host list --order=updated"
            ),
        )
        parser.add_argument(
            "--limit",
            metavar="<limit>",
            default=os.environ.get("ARA_CLI_LIMIT", 50),
            help=("Returns the first <limit> determined by the ordering. Defaults to ARA_CLI_LIMIT or 50.")
        )
        # fmt: on
        return parser

    def take_action(self, args):
        client = get_client(
            client=args.client,
            endpoint=args.server,
            timeout=args.timeout,
            username=args.username,
            password=args.password,
            verify=False if args.insecure else True,
            run_sql_migrations=False,
        )
        query = {}
        if args.name is not None:
            query["name"] = args.name

        if args.playbook is not None:
            query["playbook"] = args.playbook

        if args.with_changed:
            query["changed__gt"] = 0
        if args.without_changed:
            query["changed__lt"] = 1
        if args.with_failed:
            query["failed__gt"] = 0
        if args.without_failed:
            query["failed__lt"] = 1
        if args.with_unreachable:
            query["unreachable__gt"] = 0
        if args.without_unreachable:
            query["unreachable__lt"] = 1

        query["order"] = args.order
        query["limit"] = args.limit

        hosts = client.get("/api/v1/hosts", **query)

        if args.resolve:
            for host in hosts["results"]:
                playbook = cli_utils.get_playbook(client, host["playbook"])
                # Paths can easily take up too much width real estate
                if not args.long:
                    host["playbook"] = "(%s) %s" % (playbook["id"], cli_utils.truncatepath(playbook["path"], 50))
                else:
                    host["playbook"] = "(%s) %s" % (playbook["id"], playbook["path"])

        columns = ("id", "name", "playbook", "changed", "failed", "ok", "skipped", "unreachable", "updated")
        # fmt: off
        return (
            columns, (
                [host[column] for column in columns]
                for host in hosts["results"]
            )
        )
        # fmt: on


class HostShow(ShowOne):
    """ Returns a detailed view of a specified host """

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(HostShow, self).get_parser(prog_name)
        parser = global_arguments(parser)
        # fmt: off
        parser.add_argument(
            "host_id",
            metavar="<host-id>",
            help="Host to show",
        )
        parser.add_argument(
            "--with-facts",
            action="store_true",
            help="Also include host facts in the response (use with '-f json' or '-f yaml')"
        )
        # fmt: on
        return parser

    def take_action(self, args):
        # TODO: Render json properly in pretty tables
        if args.with_facts and args.formatter == "table":
            self.log.warn("Rendering using default table formatter, use '-f yaml' or '-f json' for improved display.")

        client = get_client(
            client=args.client,
            endpoint=args.server,
            timeout=args.timeout,
            username=args.username,
            password=args.password,
            verify=False if args.insecure else True,
            run_sql_migrations=False,
        )

        # TODO: Improve client to be better at handling exceptions
        host = client.get("/api/v1/hosts/%s" % args.host_id)
        if "detail" in host and host["detail"] == "Not found.":
            self.log.error("Host not found: %s" % args.host_id)
            sys.exit(1)

        host["report"] = "%s/playbooks/%s.html" % (args.server, host["playbook"]["id"])
        if args.with_facts:
            # fmt: off
            columns = (
                "id",
                "report",
                "name",
                "changed",
                "failed",
                "ok",
                "skipped",
                "unreachable",
                "facts",
                "updated"
            )
            # fmt: on
        else:
            # fmt: off
            columns = (
                "id",
                "report",
                "name",
                "changed",
                "failed",
                "ok",
                "skipped",
                "unreachable",
                "updated"
            )
            # fmt: on
        return (columns, ([host[column] for column in columns]))


class HostDelete(Command):
    """ Deletes the specified host and associated resources """

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(HostDelete, self).get_parser(prog_name)
        parser = global_arguments(parser)
        # fmt: off
        parser.add_argument(
            "host_id",
            metavar="<host-id>",
            help="Host to delete",
        )
        # fmt: on
        return parser

    def take_action(self, args):
        client = get_client(
            client=args.client,
            endpoint=args.server,
            timeout=args.timeout,
            username=args.username,
            password=args.password,
            verify=False if args.insecure else True,
            run_sql_migrations=False,
        )

        # TODO: Improve client to be better at handling exceptions
        client.delete("/api/v1/hosts/%s" % args.host_id)


class HostMetrics(Lister):
    """ Provides metrics about hosts """

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(HostMetrics, self).get_parser(prog_name)
        parser = global_arguments(parser)
        # fmt: off
        # Host search arguments and ordering as per ara.api.filters.HostFilter
        # TODO: non-exhaustive (searching for failed, ok, unreachable, etc.)
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=("Filter for hosts matching the provided name (full or partial)"),
        )
        parser.add_argument(
            "--playbook",
            metavar="<playbook_id>",
            default=None,
            help=("Filter for hosts for a specified playbook id"),
        )

        changed = parser.add_mutually_exclusive_group()
        changed.add_argument(
            "--with-changed",
            action="store_true",
            default=False,
            help=("Filter for hosts with changed results")
        )
        changed.add_argument(
            "--without-changed",
            action="store_true",
            default=False,
            help=("Filter out hosts without changed results")
        )

        failed = parser.add_mutually_exclusive_group()
        failed.add_argument(
            "--with-failed",
            action="store_true",
            default=False,
            help=("Filter for hosts with failed results")
        )
        failed.add_argument(
            "--without-failed",
            action="store_true",
            default=False,
            help=("Filter out hosts without failed results")
        )

        unreachable = parser.add_mutually_exclusive_group()
        unreachable.add_argument(
            "--with-unreachable",
            action="store_true",
            default=False,
            help=("Filter for hosts with unreachable results")
        )
        unreachable.add_argument(
            "--without-unreachable",
            action="store_true",
            default=False,
            help=("Filter out hosts without unreachable results")
        )
        parser.add_argument(
            "--order",
            metavar="<order>",
            default="-updated",
            help=(
                "Orders hosts by a field ('id', 'created', 'updated', 'name')\n"
                "Defaults to '-updated' descending so the most recent host is at the top.\n"
                "The order can be reversed by omitting the '-': ara host list --order=updated\n"
                "This influences the API request, not the ordering of the metrics."
            ),
        )
        parser.add_argument(
            "--limit",
            metavar="<limit>",
            default=os.environ.get("ARA_CLI_LIMIT", 1000),
            help=("Return metrics for the first <limit> determined by the ordering. Defaults to ARA_CLI_LIMIT or 1000.")
        )
        # fmt: on
        return parser

    def take_action(self, args):
        client = get_client(
            client=args.client,
            endpoint=args.server,
            timeout=args.timeout,
            username=args.username,
            password=args.password,
            verify=False if args.insecure else True,
            run_sql_migrations=False,
        )
        query = {}
        if args.name is not None:
            query["name"] = args.name

        if args.playbook is not None:
            query["playbook"] = args.playbook

        if args.with_changed:
            query["changed__gt"] = 0
        if args.without_changed:
            query["changed__lt"] = 1
        if args.with_failed:
            query["failed__gt"] = 0
        if args.without_failed:
            query["failed__lt"] = 1
        if args.with_unreachable:
            query["unreachable__gt"] = 0
        if args.without_unreachable:
            query["unreachable__lt"] = 1

        query["order"] = args.order
        query["limit"] = args.limit

        resp = client.get("/api/v1/hosts", **query)

        # Group hosts by name
        hosts = {}
        for host in resp["results"]:
            name = host["name"]
            if name not in hosts:
                hosts[name] = []
            hosts[name].append(host)

        data = {}
        for name, host_results in hosts.items():
            data[name] = {
                "name": name,
                "count": len(host_results),
                "changed": 0,
                "failed": 0,
                "ok": 0,
                "skipped": 0,
                "unreachable": 0,
            }

            for host in host_results:
                for status in ["changed", "failed", "ok", "skipped", "unreachable"]:
                    data[name][status] += host[status]

        columns = ("name", "count", "changed", "failed", "ok", "skipped", "unreachable")
        # fmt: off
        return (
            columns, (
                [data[host][column] for column in columns]
                for host in sorted(data.keys())
            )
        )
        # fmt: on



#  Copyright (c) 2018 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from rest_framework.test import APITestCase

from ara.api import models, serializers
from ara.api.tests import factories, utils


class FileTestCase(APITestCase):
    def test_file_factory(self):
        file_content = factories.FileContentFactory()
        file = factories.FileFactory(path="/path/playbook.yml", content=file_content)
        self.assertEqual(file.path, "/path/playbook.yml")
        self.assertEqual(file.content.sha1, file_content.sha1)

    def test_file_serializer(self):
        playbook = factories.PlaybookFactory()
        serializer = serializers.FileSerializer(
            data={"path": "/path/playbook.yml", "content": factories.FILE_CONTENTS, "playbook": playbook.id}
        )
        serializer.is_valid()
        file = serializer.save()
        file.refresh_from_db()
        self.assertEqual(file.content.sha1, utils.sha1(factories.FILE_CONTENTS))

    def test_create_file_with_same_content_create_only_one_file_content(self):
        playbook = factories.PlaybookFactory()
        serializer = serializers.FileSerializer(
            data={"path": "/path/1/playbook.yml", "content": factories.FILE_CONTENTS, "playbook": playbook.id}
        )
        serializer.is_valid()
        file_content = serializer.save()
        file_content.refresh_from_db()

        serializer2 = serializers.FileSerializer(
            data={"path": "/path/2/playbook.yml", "content": factories.FILE_CONTENTS, "playbook": playbook.id}
        )
        serializer2.is_valid()
        file_content = serializer2.save()
        file_content.refresh_from_db()

        self.assertEqual(2, models.File.objects.all().count())
        self.assertEqual(1, models.FileContent.objects.all().count())

    def test_create_file(self):
        self.assertEqual(0, models.File.objects.count())
        playbook = factories.PlaybookFactory()
        request = self.client.post(
            "/api/v1/files", {"path": "/path/playbook.yml", "content": factories.FILE_CONTENTS, "playbook": playbook.id}
        )
        self.assertEqual(201, request.status_code)
        self.assertEqual(1, models.File.objects.count())

    def test_post_same_file_for_a_playbook(self):
        playbook = factories.PlaybookFactory()
        self.assertEqual(0, models.File.objects.count())
        request = self.client.post(
            "/api/v1/files", {"path": "/path/playbook.yml", "content": factories.FILE_CONTENTS, "playbook": playbook.id}
        )
        self.assertEqual(201, request.status_code)
        self.assertEqual(1, models.File.objects.count())

        request = self.client.post(
            "/api/v1/files", {"path": "/path/playbook.yml", "content": factories.FILE_CONTENTS, "playbook": playbook.id}
        )
        self.assertEqual(201, request.status_code)
        self.assertEqual(1, models.File.objects.count())

    def test_get_no_files(self):
        request = self.client.get("/api/v1/files")
        self.assertEqual(0, len(request.data["results"]))

    def test_get_files(self):
        file = factories.FileFactory()
        request = self.client.get("/api/v1/files")
        self.assertEqual(1, len(request.data["results"]))
        self.assertEqual(file.path, request.data["results"][0]["path"])

    def test_get_file(self):
        file = factories.FileFactory()
        request = self.client.get("/api/v1/files/%s" % file.id)
        self.assertEqual(file.path, request.data["path"])
        self.assertEqual(file.content.sha1, request.data["sha1"])

    def test_update_file(self):
        playbook = factories.PlaybookFactory()
        file = factories.FileFactory(playbook=playbook)
        old_sha1 = file.content.sha1
        self.assertNotEqual("/path/new_playbook.yml", file.path)
        request = self.client.put(
            "/api/v1/files/%s" % file.id,
            {"path": "/path/new_playbook.yml", "content": "# playbook", "playbook": playbook.id},
        )
        self.assertEqual(200, request.status_code)
        file_updated = models.File.objects.get(id=file.id)
        self.assertEqual("/path/new_playbook.yml", file_updated.path)
        self.assertNotEqual(old_sha1, file_updated.content.sha1)

    def test_partial_update_file(self):
        file = factories.FileFactory()
        self.assertNotEqual("/path/new_playbook.yml", file.path)
        request = self.client.patch("/api/v1/files/%s" % file.id, {"path": "/path/new_playbook.yml"})
        self.assertEqual(200, request.status_code)
        file_updated = models.File.objects.get(id=file.id)
        self.assertEqual("/path/new_playbook.yml", file_updated.path)

    def test_delete_file(self):
        file = factories.FileFactory()
        self.assertEqual(1, models.File.objects.all().count())
        request = self.client.delete("/api/v1/files/%s" % file.id)
        self.assertEqual(204, request.status_code)
        self.assertEqual(0, models.File.objects.all().count())

    def test_get_file_by_date(self):
        file = factories.FileFactory()

        past = datetime.datetime.now() - datetime.timedelta(hours=12)
        negative_date_fields = ["created_before", "updated_before"]
        positive_date_fields = ["created_after", "updated_after"]

        # Expect no file when searching before it was created
        for field in negative_date_fields:
            request = self.client.get("/api/v1/files?%s=%s" % (field, past.isoformat()))
            self.assertEqual(request.data["count"], 0)

        # Expect a file when searching after it was created
        for field in positive_date_fields:
            request = self.client.get("/api/v1/files?%s=%s" % (field, past.isoformat()))
            self.assertEqual(request.data["count"], 1)
            self.assertEqual(request.data["results"][0]["id"], file.id)

    def test_get_file_order(self):
        first_file = factories.FileFactory(path="/root/file.yaml")
        second_file = factories.FileFactory(path="/root/some/path/file.yaml")

        # Ensure we have two objects
        request = self.client.get("/api/v1/files")
        self.assertEqual(2, len(request.data["results"]))

        order_fields = ["id", "created", "updated", "path"]
        # Ascending order
        for field in order_fields:
            request = self.client.get("/api/v1/files?order=%s" % field)
            self.assertEqual(request.data["results"][0]["id"], first_file.id)

        # Descending order
        for field in order_fields:
            request = self.client.get("/api/v1/files?order=-%s" % field)
            self.assertEqual(request.data["results"][0]["id"], second_file.id)

    def test_get_file_by_path(self):
        # Create two files with similar paths
        first_file = factories.FileFactory(path="/root/file.yaml")
        factories.FileFactory(path="/root/some/path/file.yaml")

        # Exact search should match one
        request = self.client.get("/api/v1/files?path=/root/file.yaml")
        self.assertEqual(1, len(request.data["results"]))
        self.assertEqual(first_file.path, request.data["results"][0]["path"])

        # Partial match should match both files
        request = self.client.get("/api/v1/files?path=file.yaml")
        self.assertEqual(2, len(request.data["results"]))

    def test_get_playbook_arguments(self):
        file = factories.FileFactory()
        request = self.client.get("/api/v1/files/%s" % file.id)
        self.assertIn("inventory", request.data["playbook"]["arguments"])



import datetime
from unittest import skip

from django.contrib.auth.models import User
from django.core.management import call_command
from django.test import LiveServerTestCase, TestCase, override_settings

from ara.api import models
from ara.api.tests import factories


class LogCheckerMixin(object):
    def run_prune_command(self, *args, **opts):
        # the command uses logging instead of prints so we need to use assertLogs
        # to retrieve and test the output
        with self.assertLogs("ara.api.management.commands.prune", "INFO") as logs:
            call_command("prune", *args, **opts)
            return logs.output


class PruneTestCase(TestCase, LogCheckerMixin):
    @skip("TODO: Why aren't logs captured properly for this test ?")
    def test_prune_without_playbooks_and_confirm(self):
        output = self.run_prune_command()
        self.assertIn(
            "INFO:ara.api.management.commands.prune:--confirm was not specified, no playbooks will be deleted", output
        )
        self.assertIn("INFO:ara.api.management.commands.prune:Found 0 playbooks matching query", output)
        self.assertIn("INFO:ara.api.management.commands.prune:0 playbooks deleted", output)

    @skip("TODO: Why aren't logs captured properly for this test ?")
    def test_prune_without_playbooks(self):
        args = ["--confirm"]
        output = self.run_prune_command(*args)
        self.assertNotIn(
            "INFO:ara.api.management.commands.prune:--confirm was not specified, no playbooks will be deleted", output
        )
        self.assertIn("INFO:ara.api.management.commands.prune:Found 0 playbooks matching query", output)
        self.assertIn("INFO:ara.api.management.commands.prune:0 playbooks deleted", output)


class PruneCmdTestCase(LiveServerTestCase, LogCheckerMixin):
    @skip("TODO: Why aren't logs captured properly for this test ?")
    def test_prune_with_no_matching_playbook(self):
        # Create a playbook with start date as of now
        factories.PlaybookFactory()
        self.assertEqual(1, models.Playbook.objects.all().count())

        args = ["--confirm"]
        output = self.run_prune_command(*args)
        self.assertIn("INFO:ara.api.management.commands.prune:Found 0 playbooks matching query", output)
        self.assertIn("INFO:ara.api.management.commands.prune:0 playbooks deleted", output)
        self.assertEqual(1, models.Playbook.objects.all().count())

    @skip("TODO: Why aren't logs captured properly for this test ?")
    def test_prune_with_matching_playbook(self):
        # Create a playbook with an old start date
        old_timestamp = datetime.datetime.now() - datetime.timedelta(days=60)
        factories.PlaybookFactory(started=old_timestamp)
        self.assertEqual(1, models.Playbook.objects.all().count())

        args = ["--confirm"]
        output = self.run_prune_command(*args)
        self.assertIn("INFO:ara.api.management.commands.prune:Found 1 playbooks matching query", output)
        self.assertIn("INFO:ara.api.management.commands.prune:1 playbooks deleted", output)
        self.assertEqual(0, models.Playbook.objects.all().count())

    def test_prune_with_no_matching_playbook_with_http_client(self):
        # Create a playbook with start date as of now
        factories.PlaybookFactory()
        self.assertEqual(1, models.Playbook.objects.all().count())

        args = ["--confirm", "--client", "http", "--endpoint", self.live_server_url]
        output = self.run_prune_command(*args)
        self.assertIn("INFO:ara.api.management.commands.prune:Found 0 playbooks matching query", output)
        self.assertIn("INFO:ara.api.management.commands.prune:0 playbooks deleted", output)
        self.assertEqual(1, models.Playbook.objects.all().count())

    def test_prune_with_matching_playbook_with_http_client(self):
        # Create a playbook with an old start date
        old_timestamp = datetime.datetime.now() - datetime.timedelta(days=60)
        factories.PlaybookFactory(started=old_timestamp)
        self.assertEqual(1, models.Playbook.objects.all().count())

        args = ["--confirm", "--client", "http", "--endpoint", self.live_server_url]
        output = self.run_prune_command(*args)
        self.assertIn("INFO:ara.api.management.commands.prune:Found 1 playbooks matching query", output)
        self.assertIn("INFO:ara.api.management.commands.prune:1 playbooks deleted", output)
        self.assertEqual(0, models.Playbook.objects.all().count())

    @override_settings(READ_LOGIN_REQUIRED=True, WRITE_LOGIN_REQUIRED=True)
    def test_prune_without_authenticated_http_client(self):
        args = ["--confirm", "--client", "http", "--endpoint", self.live_server_url]
        with self.assertRaises(SystemExit):
            self.run_prune_command(*args)

    @override_settings(READ_LOGIN_REQUIRED=True, WRITE_LOGIN_REQUIRED=True)
    def test_prune_with_authenticated_http_client(self):
        # Create a user
        self.user = User.objects.create_superuser("prune", "prune@example.org", "password")

        # Create a playbook with an old start date
        old_timestamp = datetime.datetime.now() - datetime.timedelta(days=60)
        factories.PlaybookFactory(started=old_timestamp)
        self.assertEqual(1, models.Playbook.objects.all().count())

        args = [
            "--confirm",
            "--client",
            "http",
            "--endpoint",
            self.live_server_url,
            "--username",
            "prune",
            "--password",
            "password",
        ]
        output = self.run_prune_command(*args)
        self.assertIn("INFO:ara.api.management.commands.prune:Found 1 playbooks matching query", output)
        self.assertIn("INFO:ara.api.management.commands.prune:1 playbooks deleted", output)
        self.assertEqual(0, models.Playbook.objects.all().count())

    @override_settings(READ_LOGIN_REQUIRED=True, WRITE_LOGIN_REQUIRED=True)
    def test_prune_with_bad_authentication_http_client(self):
        # Create a user
        self.user = User.objects.create_superuser("prune", "prune@example.org", "password")

        # Create a playbook with an old start date
        old_timestamp = datetime.datetime.now() - datetime.timedelta(days=60)
        factories.PlaybookFactory(started=old_timestamp)
        self.assertEqual(1, models.Playbook.objects.all().count())

        # Set up arguments with a wrong password
        args = [
            "--confirm",
            "--client",
            "http",
            "--endpoint",
            self.live_server_url,
            "--username",
            "prune",
            "--password",
            "somethingelse",
        ]

        with self.assertRaises(SystemExit):
            self.run_prune_command(*args)
            # TODO: the assertRaises prevents us from looking at the output
            # output = run_prune_command(*args)
            # self.assertIn("Client failed to retrieve results, see logs for ara.clients.offline or ara.clients.http.", output)  # noqa

        # Nothing should have been deleted because the command failed
        self.assertEqual(1, models.Playbook.objects.all().count())



#  Copyright (c) 2018 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from django.utils import timezone
from django.utils.dateparse import parse_duration
from rest_framework.test import APITestCase

from ara.api import models, serializers
from ara.api.tests import factories, utils


class ResultTestCase(APITestCase):
    def test_result_factory(self):
        result = factories.ResultFactory(status="failed")
        self.assertEqual(result.status, "failed")

    def test_result_serializer(self):
        host = factories.HostFactory()
        task = factories.TaskFactory()
        serializer = serializers.ResultSerializer(
            data={
                "status": "skipped",
                "host": host.id,
                "task": task.id,
                "play": task.play.id,
                "playbook": task.playbook.id,
                "changed": False,
                "ignore_errors": False,
            }
        )
        serializer.is_valid()
        result = serializer.save()
        result.refresh_from_db()
        self.assertEqual(result.status, "skipped")
        self.assertEqual(result.changed, False)
        self.assertEqual(result.ignore_errors, False)
        self.assertEqual(result.host.id, host.id)
        self.assertEqual(result.task.id, task.id)

    def test_result_serializer_compress_content(self):
        host = factories.HostFactory()
        task = factories.TaskFactory()
        serializer = serializers.ResultSerializer(
            data={
                "content": factories.RESULT_CONTENTS,
                "status": "ok",
                "host": host.id,
                "task": task.id,
                "play": task.play.id,
                "playbook": task.playbook.id,
            }
        )
        serializer.is_valid()
        result = serializer.save()
        result.refresh_from_db()
        self.assertEqual(result.content, utils.compressed_obj(factories.RESULT_CONTENTS))

    def test_result_serializer_decompress_content(self):
        result = factories.ResultFactory(content=utils.compressed_obj(factories.RESULT_CONTENTS))
        serializer = serializers.ResultSerializer(instance=result)
        self.assertEqual(serializer.data["content"], factories.RESULT_CONTENTS)

    def test_get_no_results(self):
        request = self.client.get("/api/v1/results")
        self.assertEqual(0, len(request.data["results"]))

    def test_get_results(self):
        result = factories.ResultFactory()
        request = self.client.get("/api/v1/results")
        self.assertEqual(1, len(request.data["results"]))
        self.assertEqual(result.status, request.data["results"][0]["status"])

    def test_delete_result(self):
        result = factories.ResultFactory()
        self.assertEqual(1, models.Result.objects.all().count())
        request = self.client.delete("/api/v1/results/%s" % result.id)
        self.assertEqual(204, request.status_code)
        self.assertEqual(0, models.Result.objects.all().count())

    def test_create_result(self):
        host = factories.HostFactory()
        task = factories.TaskFactory()
        self.assertEqual(0, models.Result.objects.count())
        request = self.client.post(
            "/api/v1/results",
            {
                "content": factories.RESULT_CONTENTS,
                "status": "ok",
                "host": host.id,
                "task": task.id,
                "play": task.play.id,
                "playbook": task.playbook.id,
                "changed": True,
                "ignore_errors": False,
            },
        )
        self.assertEqual(201, request.status_code)
        self.assertEqual(request.data["changed"], True)
        self.assertEqual(request.data["ignore_errors"], False)
        self.assertEqual(1, models.Result.objects.count())

    def test_partial_update_result(self):
        result = factories.ResultFactory()
        self.assertNotEqual("unreachable", result.status)
        request = self.client.patch("/api/v1/results/%s" % result.id, {"status": "unreachable"})
        self.assertEqual(200, request.status_code)
        result_updated = models.Result.objects.get(id=result.id)
        self.assertEqual("unreachable", result_updated.status)

    def test_get_result(self):
        result = factories.ResultFactory()
        request = self.client.get("/api/v1/results/%s" % result.id)
        self.assertEqual(result.status, request.data["status"])

    def test_get_result_by_association(self):
        # Create two results in necessarily two different playbooks with different children:
        # playbook -> play -> task -> result <- host
        first_result = factories.ResultFactory()
        second_result = factories.ResultFactory()

        # the fields with the association ids
        associations = ["playbook", "play", "task", "host"]

        # Validate that we somehow didn't wind up with the same association ids
        for association in associations:
            first = getattr(first_result, association)
            second = getattr(second_result, association)
            self.assertNotEqual(first.id, second.id)

        # In other words, there must be two distinct results
        request = self.client.get("/api/v1/results")
        self.assertEqual(2, request.data["count"])
        self.assertEqual(2, len(request.data["results"]))

        # Searching for the first_result associations should only yield one result
        for association in associations:
            assoc_id = getattr(first_result, association).id
            results = self.client.get("/api/v1/results?%s=%s" % (association, assoc_id))
            self.assertEqual(1, results.data["count"])
            self.assertEqual(1, len(results.data["results"]))
            self.assertEqual(assoc_id, results.data["results"][0][association])

    def test_get_result_by_statuses(self):
        failed_result = factories.ResultFactory(status="failed")
        skipped_result = factories.ResultFactory(status="skipped")
        factories.ResultFactory(status="ok")
        results = self.client.get("/api/v1/results").data["results"]
        self.assertEqual(3, len(results))

        results = self.client.get("/api/v1/results?status=failed").data["results"]
        self.assertEqual(1, len(results))
        self.assertEqual(failed_result.status, results[0]["status"])

        results = self.client.get("/api/v1/results?status=skipped").data["results"]
        self.assertEqual(1, len(results))
        self.assertEqual(skipped_result.status, results[0]["status"])

        results = self.client.get("/api/v1/results?status=failed&status=skipped").data["results"]
        self.assertEqual(2, len(results))
        self.assertEqual(failed_result.status, results[1]["status"])
        self.assertEqual(skipped_result.status, results[0]["status"])

    def test_result_status_serializer(self):
        ok = factories.ResultFactory(status="ok")
        result = self.client.get("/api/v1/results/%s" % ok.id)
        self.assertEqual(result.data["status"], "ok")

        changed = factories.ResultFactory(status="ok", changed=True)
        result = self.client.get("/api/v1/results/%s" % changed.id)
        self.assertEqual(result.data["status"], "changed")

        failed = factories.ResultFactory(status="failed")
        result = self.client.get("/api/v1/results/%s" % failed.id)
        self.assertEqual(result.data["status"], "failed")

        ignored = factories.ResultFactory(status="failed", ignore_errors=True)
        result = self.client.get("/api/v1/results/%s" % ignored.id)
        self.assertEqual(result.data["status"], "ignored")

        skipped = factories.ResultFactory(status="skipped")
        result = self.client.get("/api/v1/results/%s" % skipped.id)
        self.assertEqual(result.data["status"], "skipped")

        unreachable = factories.ResultFactory(status="unreachable")
        result = self.client.get("/api/v1/results/%s" % unreachable.id)
        self.assertEqual(result.data["status"], "unreachable")

    def test_get_result_with_ignore_errors(self):
        failed = factories.ResultFactory(status="failed", ignore_errors=False)
        ignored = factories.ResultFactory(status="failed", ignore_errors=True)

        # Searching for failed should return both
        results = self.client.get("/api/v1/results?status=failed").data["results"]
        self.assertEqual(2, len(results))

        # Searching for failed with ignore_errors=True should only return the ignored result
        results = self.client.get("/api/v1/results?status=failed&ignore_errors=true").data["results"]
        self.assertEqual(1, len(results))
        self.assertEqual(ignored.id, results[0]["id"])

        # Searching for failed with ignore_errors=False should only return the failed result
        results = self.client.get("/api/v1/results?status=failed&ignore_errors=false").data["results"]
        self.assertEqual(1, len(results))
        self.assertEqual(failed.id, results[0]["id"])

    def test_get_result_duration(self):
        started = timezone.now()
        ended = started + datetime.timedelta(hours=1)
        result = factories.ResultFactory(started=started, ended=ended)
        request = self.client.get("/api/v1/results/%s" % result.id)
        self.assertEqual(parse_duration(request.data["duration"]), ended - started)

    def test_get_result_by_date(self):
        result = factories.ResultFactory()

        past = datetime.datetime.now() - datetime.timedelta(hours=12)
        negative_date_fields = ["created_before", "started_before", "updated_before"]
        positive_date_fields = ["created_after", "started_after", "updated_after"]

        # Expect no result when searching before it was created
        for field in negative_date_fields:
            request = self.client.get("/api/v1/results?%s=%s" % (field, past.isoformat()))
            self.assertEqual(request.data["count"], 0)

        # Expect a result when searching after it was created
        for field in positive_date_fields:
            request = self.client.get("/api/v1/results?%s=%s" % (field, past.isoformat()))
            self.assertEqual(request.data["count"], 1)
            self.assertEqual(request.data["results"][0]["id"], result.id)

    def test_get_result_order(self):
        old_started = timezone.now() - datetime.timedelta(hours=12)
        old_ended = old_started + datetime.timedelta(minutes=30)
        old_result = factories.ResultFactory(started=old_started, ended=old_ended)
        new_started = timezone.now() - datetime.timedelta(hours=6)
        new_ended = new_started + datetime.timedelta(hours=1)
        new_result = factories.ResultFactory(started=new_started, ended=new_ended)

        # Ensure we have two objects
        request = self.client.get("/api/v1/results")
        self.assertEqual(2, len(request.data["results"]))

        order_fields = ["id", "created", "updated", "started", "ended", "duration"]
        # Ascending order
        for field in order_fields:
            request = self.client.get("/api/v1/results?order=%s" % field)
            self.assertEqual(request.data["results"][0]["id"], old_result.id)

        # Descending order
        for field in order_fields:
            request = self.client.get("/api/v1/results?order=-%s" % field)
            self.assertEqual(request.data["results"][0]["id"], new_result.id)

    def test_get_changed_results(self):
        changed_result = factories.ResultFactory(changed=True)
        unchanged_result = factories.ResultFactory(changed=False)

        # Assert two results
        results = self.client.get("/api/v1/results").data["results"]
        self.assertEqual(2, len(results))

        # Assert one changed
        results = self.client.get("/api/v1/results?changed=true").data["results"]
        self.assertEqual(1, len(results))
        self.assertEqual(results[0]["id"], changed_result.id)

        # Assert one unchanged
        results = self.client.get("/api/v1/results?changed=false").data["results"]
        self.assertEqual(1, len(results))
        self.assertEqual(results[0]["id"], unchanged_result.id)

    def test_get_playbook_arguments(self):
        result = factories.ResultFactory()
        request = self.client.get("/api/v1/results/%s" % result.id)
        self.assertIn("inventory", request.data["playbook"]["arguments"])



#  Copyright (c) 2018 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

from rest_framework.routers import DefaultRouter

from ara.api import views

router = DefaultRouter(trailing_slash=False)
router.register("labels", views.LabelViewSet, basename="label")
router.register("playbooks", views.PlaybookViewSet, basename="playbook")
router.register("plays", views.PlayViewSet, basename="play")
router.register("tasks", views.TaskViewSet, basename="task")
router.register("hosts", views.HostViewSet, basename="host")
router.register("results", views.ResultViewSet, basename="result")
router.register("files", views.FileViewSet, basename="file")
router.register("records", views.RecordViewSet, basename="record")

urlpatterns = router.urls



#  Copyright (c) 2018 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

from rest_framework.test import APITestCase

from ara.api.tests import factories


class FileContentTestCase(APITestCase):
    def test_file_content_factory(self):
        file_content = factories.FileContentFactory(sha1="413a2f16b8689267b7d0c2e10cdd19bf3e54208d")
        self.assertEqual(file_content.sha1, "413a2f16b8689267b7d0c2e10cdd19bf3e54208d")



#  Copyright (c) 2018 Red Hat, Inc.
#
#  This file is part of ARA: Ansible Run Analysis.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

from ansible.playbook.play import Play
from ansible.plugins.action import ActionBase

from ara.clients import utils as client_utils

DOCUMENTATION = """
---
module: ara_record
short_description: Ansible module to record persistent data with ARA.
version_added: "2.0"
author: "David Moreau-Simard <dmsimard@redhat.com>"
description:
    - Ansible module to record persistent data with ARA.
options:
    playbook_id:
        description:
            - id of the playbook to write the key to
            - if not set, the module will use the ongoing playbook's id
        required: false
    key:
        description:
            - Name of the key to write data to
        required: true
    value:
        description:
            - Value of the key written to
        required: true
    type:
        description:
            - Type of the key
        choices: [text, url, json, list, dict]
        default: text

requirements:
    - "python >= 3.5"
    - "ara >= 1.0.0"
"""

EXAMPLES = """
- name: Associate specific data to a key for a playbook
  ara_record:
    key: "foo"
    value: "bar"

- name: Associate data to a playbook that previously ran
  ara_record:
    playbook_id: 21
    key: logs
    value: "{{ lookup('file', '/var/log/ansible.log') }}"
    type: text

- name: Retrieve the git version of the development repository
  shell: cd dev && git rev-parse HEAD
  register: git_version
  delegate_to: localhost

- name: Record and register the git version of the playbooks
  ara_record:
    key: "git_version"
    value: "{{ git_version.stdout }}"
  register: version

- name: Print recorded data
  debug:
    msg: "{{ version.playbook_id }} - {{ version.key }}: {{ version.value }}

# Write data with a type (otherwise defaults to "text")
# This changes the behavior on how the value is presented in the web interface
- name: Record different formats of things
  ara_record:
    key: "{{ item.key }}"
    value: "{{ item.value }}"
    type: "{{ item.type }}"
  with_items:
    - { key: "log", value: "error", type: "text" }
    - { key: "website", value: "http://domain.tld", type: "url" }
    - { key: "data", value: "{ 'key': 'value' }", type: "json" }
    - { key: "somelist", value: ['one', 'two'], type: "list" }
    - { key: "somedict", value: {'key': 'value' }, type: "dict" }
"""


RETURN = """
playbook:
    description: ID of the playbook the data was recorded in
    returned: on success
    type: int
    sample: 1
key:
    description: Key where the record is saved
    returned: on success
    type: str
    sample: log_url
value:
    description: Value of the key
    returned: on success
    type: complex
    sample: http://logurl
type:
    description: Type of the key
    returned: on success
    type: string
    sample: url
created:
    description: Date the record was created (ISO-8601)
    returned: on success
    type: str
    sample: 2018-11-15T17:27:41.597234Z
updated:
    description: Date the record was updated (ISO-8601)
    returned: on success
    type: str
    sample: 2018-11-15T17:27:41.597265Z
"""


class ActionModule(ActionBase):
    """ Record persistent data as key/value pairs in ARA """

    TRANSFERS_FILES = False
    VALID_ARGS = frozenset(("playbook_id", "key", "value", "type"))
    VALID_TYPES = ["text", "url", "json", "list", "dict"]

    def __init__(self, *args, **kwargs):
        super(ActionModule, self).__init__(*args, **kwargs)
        self.client = client_utils.active_client()

    def create_or_update_key(self, playbook, key, value, type):
        changed = False
        record = self.client.get("/api/v1/records?playbook=%s&key=%s" % (playbook, key))
        if record["count"] == 0:
            # Create the record if it doesn't exist
            record = self.client.post("/api/v1/records", playbook=playbook, key=key, value=value, type=type)
            changed = True
        else:
            # Otherwise update it if the data is different (idempotency)
            old = self.client.get("/api/v1/records/%s" % record["results"][0]["id"])
            if old["value"] != value or old["type"] != type:
                record = self.client.patch("/api/v1/records/%s" % old["id"], key=key, value=value, type=type)
                changed = True
            else:
                record = old
        return record, changed

    def run(self, tmp=None, task_vars=None):
        if task_vars is None:
            task_vars = dict()

        for arg in self._task.args:
            if arg not in self.VALID_ARGS:
                result = {"failed": True, "msg": "{0} is not a valid option.".format(arg)}
                return result

        result = super(ActionModule, self).run(tmp, task_vars)

        playbook_id = self._task.args.get("playbook_id", None)
        key = self._task.args.get("key", None)
        value = self._task.args.get("value", None)
        type = self._task.args.get("type", "text")

        required = ["key", "value"]
        for parameter in required:
            if not self._task.args.get(parameter):
                result["failed"] = True
                result["msg"] = "Parameter '{0}' is required".format(parameter)
                return result

        if type not in self.VALID_TYPES:
            result["failed"] = True
            msg = "Type '{0}' is not supported, choose one of: {1}".format(type, ", ".join(self.VALID_TYPES))
            result["msg"] = msg
            return result

        if playbook_id is None:
            # Retrieve the playbook id by working our way up from the task to find
            # the play uuid. Once we have the play uuid, we can find the playbook.
            parent = self._task
            while not isinstance(parent._parent._play, Play):
                parent = parent._parent

            play = self.client.get("/api/v1/plays?uuid=%s" % parent._parent._play._uuid)
            playbook_id = play["results"][0]["playbook"]

        try:
            data, changed = self.create_or_update_key(playbook_id, key, value, type)
            result["changed"] = changed
            result["key"] = data["key"]
            result["value"] = data["value"]
            result["type"] = data["type"]
            result["playbook_id"] = data["playbook"]
            result["created"] = data["created"]
            result["updated"] = data["updated"]
            if result["changed"]:
                result["msg"] = "Record created or updated in ARA"
            else:
                result["msg"] = "Record unchanged in ARA"
        except Exception as e:
            result["changed"] = False
            result["failed"] = True
            result["msg"] = "Record failed to be created or updated in ARA: {0}".format(str(e))
        return result



#  Copyright (c) 2018 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

from rest_framework import serializers

from ara.api import fields as ara_fields, models


class ResultStatusSerializer(serializers.ModelSerializer):
    class Meta:
        abstract = True

    status = serializers.SerializerMethodField()

    @staticmethod
    def get_status(obj):
        if obj.status == "ok" and obj.changed:
            return "changed"
        elif obj.status == "failed" and obj.ignore_errors:
            return "ignored"
        else:
            return obj.status


class TaskPathSerializer(serializers.ModelSerializer):
    class Meta:
        abstract = True

    path = serializers.SerializerMethodField()

    @staticmethod
    def get_path(obj):
        return obj.file.path


class ItemCountSerializer(serializers.ModelSerializer):
    class Meta:
        abstract = True

    # For counting relationships to other objects
    items = serializers.SerializerMethodField()

    @staticmethod
    def get_items(obj):
        types = ["plays", "tasks", "results", "hosts", "files", "records"]
        items = {item: getattr(obj, item).count() for item in types if hasattr(obj, item)}
        return items


class FileSha1Serializer(serializers.ModelSerializer):
    class Meta:
        abstract = True

    # For retrieving the sha1 of a file's contents
    sha1 = serializers.SerializerMethodField()

    @staticmethod
    def get_sha1(obj):
        return obj.content.sha1


#######
# Simple serializers provide lightweight representations of objects suitable for inclusion in other objects
#######


class SimpleLabelSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Label
        exclude = ("created", "updated")


class SimplePlaybookSerializer(ItemCountSerializer):
    class Meta:
        model = models.Playbook
        exclude = ("created", "updated")

    arguments = ara_fields.CompressedObjectField(default=ara_fields.EMPTY_DICT, read_only=True)
    labels = SimpleLabelSerializer(many=True, read_only=True, default=[])


class SimplePlaySerializer(ItemCountSerializer):
    class Meta:
        model = models.Play
        exclude = ("playbook", "uuid", "created", "updated")


class SimpleTaskSerializer(ItemCountSerializer, TaskPathSerializer):
    class Meta:
        model = models.Task
        exclude = ("playbook", "play", "created", "updated")

    tags = ara_fields.CompressedObjectField(read_only=True)


class SimpleHostSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Host
        exclude = ("playbook", "facts", "created", "updated")


class SimpleFileSerializer(FileSha1Serializer):
    class Meta:
        model = models.File
        exclude = ("playbook", "content", "created", "updated")


#######
# Detailed serializers returns every field of an object as well as a simple
# representation of relationships to other objects.
#######


class DetailedLabelSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Label
        fields = "__all__"


class DetailedPlaybookSerializer(ItemCountSerializer):
    class Meta:
        model = models.Playbook
        fields = "__all__"

    arguments = ara_fields.CompressedObjectField(default=ara_fields.EMPTY_DICT, read_only=True)
    labels = SimpleLabelSerializer(many=True, read_only=True, default=[])


class DetailedPlaySerializer(ItemCountSerializer):
    class Meta:
        model = models.Play
        fields = "__all__"

    playbook = SimplePlaybookSerializer(read_only=True)


class DetailedTaskSerializer(ItemCountSerializer, TaskPathSerializer):
    class Meta:
        model = models.Task
        fields = "__all__"

    playbook = SimplePlaybookSerializer(read_only=True)
    play = SimplePlaySerializer(read_only=True)
    file = SimpleFileSerializer(read_only=True)
    tags = ara_fields.CompressedObjectField(read_only=True)


class DetailedHostSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Host
        fields = "__all__"

    playbook = SimplePlaybookSerializer(read_only=True)
    facts = ara_fields.CompressedObjectField(read_only=True)


class DetailedResultSerializer(ResultStatusSerializer):
    class Meta:
        model = models.Result
        fields = "__all__"

    playbook = SimplePlaybookSerializer(read_only=True)
    play = SimplePlaySerializer(read_only=True)
    task = SimpleTaskSerializer(read_only=True)
    host = SimpleHostSerializer(read_only=True)
    content = ara_fields.CompressedObjectField(read_only=True)


class DetailedFileSerializer(FileSha1Serializer):
    class Meta:
        model = models.File
        fields = "__all__"

    playbook = SimplePlaybookSerializer(read_only=True)
    content = ara_fields.FileContentField(read_only=True)


class DetailedRecordSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Record
        fields = "__all__"

    playbook = SimplePlaybookSerializer(read_only=True)
    value = ara_fields.CompressedObjectField(read_only=True)


#######
# List serializers returns lightweight fields about objects.
# Relationships are represented by numerical IDs.
#######


class ListLabelSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Label
        fields = "__all__"


class ListPlaybookSerializer(ItemCountSerializer):
    class Meta:
        model = models.Playbook
        fields = "__all__"

    arguments = ara_fields.CompressedObjectField(default=ara_fields.EMPTY_DICT, read_only=True)
    labels = SimpleLabelSerializer(many=True, read_only=True, default=[])


class ListPlaySerializer(ItemCountSerializer):
    class Meta:
        model = models.Play
        fields = "__all__"

    playbook = serializers.PrimaryKeyRelatedField(read_only=True)


class ListTaskSerializer(ItemCountSerializer, TaskPathSerializer):
    class Meta:
        model = models.Task
        fields = "__all__"

    tags = ara_fields.CompressedObjectField(read_only=True)
    play = serializers.PrimaryKeyRelatedField(read_only=True)


class ListHostSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Host
        exclude = ("facts",)

    playbook = serializers.PrimaryKeyRelatedField(read_only=True)


class ListResultSerializer(ResultStatusSerializer):
    class Meta:
        model = models.Result
        exclude = ("content",)

    playbook = serializers.PrimaryKeyRelatedField(read_only=True)
    play = serializers.PrimaryKeyRelatedField(read_only=True)
    task = serializers.PrimaryKeyRelatedField(read_only=True)
    host = serializers.PrimaryKeyRelatedField(read_only=True)


class ListFileSerializer(FileSha1Serializer):
    class Meta:
        model = models.File
        exclude = ("content",)

    playbook = serializers.PrimaryKeyRelatedField(read_only=True)


class ListRecordSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Record
        exclude = ("value",)

    playbook = serializers.PrimaryKeyRelatedField(read_only=True)


#######
# Default serializers represents objects as they are modelized in the database.
# They are used for creating/updating/destroying objects.
#######


class LabelSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Label
        fields = "__all__"


class PlaybookSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Playbook
        fields = "__all__"

    arguments = ara_fields.CompressedObjectField(default=ara_fields.EMPTY_DICT)
    labels = ara_fields.CreatableSlugRelatedField(
        many=True, slug_field="name", queryset=models.Label.objects.all(), required=False
    )


class PlaySerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Play
        fields = "__all__"


class TaskSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Task
        fields = "__all__"

    tags = ara_fields.CompressedObjectField(default=ara_fields.EMPTY_LIST, help_text="A list containing Ansible tags")


class HostSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Host
        fields = "__all__"

    facts = ara_fields.CompressedObjectField(default=ara_fields.EMPTY_DICT)

    def get_unique_together_validators(self):
        """
        Hosts have a "unique together" constraint for host.name and play.id.
        We want to have a "get_or_create" facility and in order to do that, we
        must manage the validation during the creation, not before.
        Overriding this method effectively disables this validator.
        """
        return []

    def create(self, validated_data):
        host, created = models.Host.objects.get_or_create(
            name=validated_data["name"], playbook=validated_data["playbook"], defaults=validated_data
        )
        return host


class ResultSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Result
        fields = "__all__"

    content = ara_fields.CompressedObjectField(default=ara_fields.EMPTY_DICT)


class FileSerializer(FileSha1Serializer):
    class Meta:
        model = models.File
        fields = "__all__"

    content = ara_fields.FileContentField()

    def get_unique_together_validators(self):
        """
        Files have a "unique together" constraint for file.path and playbook.id.
        We want to have a "get_or_create" facility and in order to do that, we
        must manage the validation during the creation, not before.
        Overriding this method effectively disables this validator.
        """
        return []

    def create(self, validated_data):
        file_, created = models.File.objects.get_or_create(
            path=validated_data["path"],
            content=validated_data["content"],
            playbook=validated_data["playbook"],
            defaults=validated_data,
        )
        return file_


class RecordSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Record
        fields = "__all__"

    value = ara_fields.CompressedObjectField(
        default=ara_fields.EMPTY_STRING, help_text="A string, list, dict, json or other formatted data"
    )



# Generated by Django 2.2.1 on 2019-05-30 16:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0002_remove_host_alias'),
    ]

    operations = [
        migrations.AddField(
            model_name='result',
            name='changed',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='result',
            name='ignore_errors',
            field=models.BooleanField(default=False),
        ),
    ]



#  Copyright (c) 2019 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.test import RequestFactory, TestCase, override_settings

from ara.api.auth import APIAccessPermission


class User:
    is_authenticated = True


class AnonymousUser(User):
    is_authenticated = False


class PermissionBackendTestCase(TestCase):
    def setUp(self):
        factory = RequestFactory()
        self.anon_get_request = factory.get("/")
        self.anon_get_request.user = AnonymousUser()
        self.anon_post_request = factory.post("/")
        self.anon_post_request.user = AnonymousUser()

        self.authed_get_request = factory.get("/")
        self.authed_get_request.user = User()
        self.authed_post_request = factory.post("/")
        self.authed_post_request.user = User()

    @override_settings(READ_LOGIN_REQUIRED=False, WRITE_LOGIN_REQUIRED=True)
    def test_anonymous_read_access(self):
        backend = APIAccessPermission()

        # Writes are blocked (just to show it has no affect on read)
        self.assertFalse(backend.has_permission(self.anon_post_request, None))

        # Reads are allowed based on READ_LOGIN_REQUIRED
        self.assertTrue(backend.has_permission(self.anon_get_request, None))
        settings.READ_LOGIN_REQUIRED = True
        self.assertFalse(backend.has_permission(self.anon_get_request, None))

    @override_settings(READ_LOGIN_REQUIRED=True, WRITE_LOGIN_REQUIRED=False)
    def test_anonymous_write_access(self):
        backend = APIAccessPermission()

        # Reads are blocked (just to show it has no affect on write)
        self.assertFalse(backend.has_permission(self.anon_get_request, None))

        # Writes are allowed based on WRITE_LOGIN_REQUIRED
        self.assertTrue(backend.has_permission(self.anon_post_request, None))
        settings.WRITE_LOGIN_REQUIRED = True
        self.assertFalse(backend.has_permission(self.anon_post_request, None))

    @override_settings(READ_LOGIN_REQUIRED=True, WRITE_LOGIN_REQUIRED=True)
    def test_auth_access(self):
        backend = APIAccessPermission()

        self.assertTrue(backend.has_permission(self.authed_get_request, None))
        self.assertTrue(backend.has_permission(self.authed_post_request, None))



#  Copyright (c) 2019 Red Hat, Inc.
#
#  This file is part of ARA Records Ansible.
#
#  ARA is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ARA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ARA.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import os
from distutils.sysconfig import get_python_lib

from . import action_plugins, callback_plugins, lookup_plugins

exports = """
export ANSIBLE_CALLBACK_PLUGINS=${{ANSIBLE_CALLBACK_PLUGINS:-}}${{ANSIBLE_CALLBACK_PLUGINS+:}}{}
export ANSIBLE_ACTION_PLUGINS=${{ANSIBLE_ACTION_PLUGINS:-}}${{ANSIBLE_ACTION_PLUGINS+:}}{}
export ANSIBLE_LOOKUP_PLUGINS=${{ANSIBLE_LOOKUP_PLUGINS:-}}${{ANSIBLE_LOOKUP_PLUGINS+:}}{}
""".format(
    callback_plugins, action_plugins, lookup_plugins
)

if "VIRTUAL_ENV" in os.environ:
    """ PYTHONPATH may be exported when 'ara' module is installed in a
    virtualenv and ansible is installed on system python to avoid ansible
    failure to find ara module.
    """
    # inspired by https://stackoverflow.com/a/122340/99834
    lib = get_python_lib()
    if "PYTHONPATH" in os.environ:
        python_paths = os.environ["PYTHONPATH"].split(os.pathsep)
    else:
        python_paths = []
    if lib not in python_paths:
        python_paths.append(lib)
        exports += "export PYTHONPATH=${PYTHONPATH:-}${PYTHONPATH+:}%s\n" % os.pathsep.join(python_paths)

if __name__ == "__main__":
    print(exports.strip())



# Copyright (c) 2020 The ARA Records Ansible authors
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import logging
import os
import sys

from cliff.command import Command
from cliff.lister import Lister
from cliff.show import ShowOne

import ara.cli.utils as cli_utils
from ara.cli.base import global_arguments
from ara.clients.utils import get_client


class TaskList(Lister):
    """ Returns a list of tasks based on search queries """

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(TaskList, self).get_parser(prog_name)
        parser = global_arguments(parser)
        # fmt: off
        # Task search arguments and ordering as per ara.api.filters.TaskFilter
        parser.add_argument(
            "--playbook",
            metavar="<playbook_id>",
            default=None,
            help=("List tasks for a specified playbook id"),
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            default=None,
            help=("List tasks matching a specific status ('completed', 'running' or 'unknown')")
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=("List tasks matching the provided name (full or partial)"),
        )
        parser.add_argument(
            "--path",
            metavar="<path>",
            default=None,
            help=("List tasks matching the provided path (full or partial)"),
        )
        parser.add_argument(
            "--action",
            metavar="<action>",
            default=None,
            help=("List tasks matching a specific action/ansible module (ex: 'debug', 'package', 'set_fact')"),
        )
        parser.add_argument(
            "--long",
            action="store_true",
            default=False,
            help=("Don't truncate paths and include additional fields: path, lineno, handler, play")
        )
        parser.add_argument(
            "--resolve",
            action="store_true",
            default=os.environ.get("ARA_CLI_RESOLVE", False),
            help=("Resolve IDs to identifiers (such as path or names). Defaults to ARA_CLI_RESOLVE or False")
        )
        parser.add_argument(
            "--order",
            metavar="<order>",
            default="-started",
            help=(
                "Orders tasks by a field ('id', 'created', 'updated', 'started', 'ended', 'duration')\n"
                "Defaults to '-started' descending so the most recent task is at the top.\n"
                "The order can be reversed by omitting the '-': ara task list --order=started"
            ),
        )
        parser.add_argument(
            "--limit",
            metavar="<limit>",
            default=os.environ.get("ARA_CLI_LIMIT", 50),
            help=("Returns the first <limit> determined by the ordering. Defaults to ARA_CLI_LIMIT or 50.")
        )
        # fmt: on
        return parser

    def take_action(self, args):
        client = get_client(
            client=args.client,
            endpoint=args.server,
            timeout=args.timeout,
            username=args.username,
            password=args.password,
            verify=False if args.insecure else True,
            run_sql_migrations=False,
        )
        query = {}
        if args.playbook is not None:
            query["playbook"] = args.playbook

        if args.status is not None:
            query["status"] = args.status

        if args.name is not None:
            query["name"] = args.name

        if args.path is not None:
            query["path"] = args.path

        if args.action is not None:
            query["action"] = args.action

        query["order"] = args.order
        query["limit"] = args.limit

        tasks = client.get("/api/v1/tasks", **query)

        for task in tasks["results"]:
            task["results"] = task["items"]["results"]
            if args.resolve:
                playbook = cli_utils.get_playbook(client, task["playbook"])
                # Paths can easily take up too much width real estate
                if not args.long:
                    task["playbook"] = "(%s) %s" % (playbook["id"], cli_utils.truncatepath(playbook["path"], 50))
                else:
                    task["playbook"] = "(%s) %s" % (playbook["id"], playbook["path"])

                if args.long:
                    play = cli_utils.get_play(client, task["play"])
                    task["play"] = "(%s) %s" % (play["id"], play["name"])

        # fmt: off
        if args.long:
            columns = (
                "id",
                "status",
                "results",
                "action",
                "name",
                "tags",
                "path",
                "lineno",
                "handler",
                "playbook",
                "play",
                "started",
                "duration"
            )
        else:
            columns = (
                "id",
                "status",
                "results",
                "action",
                "name",
                "playbook",
                "started",
                "duration"
            )
        # fmt: off
        return (
            columns, (
                [task[column] for column in columns]
                for task in tasks["results"]
            )
        )
        # fmt: on


class TaskShow(ShowOne):
    """ Returns a detailed view of a specified task """

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(TaskShow, self).get_parser(prog_name)
        parser = global_arguments(parser)
        # fmt: off
        parser.add_argument(
            "task_id",
            metavar="<task-id>",
            help="Task to show",
        )
        # fmt: on
        return parser

    def take_action(self, args):
        client = get_client(
            client=args.client,
            endpoint=args.server,
            timeout=args.timeout,
            username=args.username,
            password=args.password,
            verify=False if args.insecure else True,
            run_sql_migrations=False,
        )

        # TODO: Improve client to be better at handling exceptions
        task = client.get("/api/v1/tasks/%s" % args.task_id)
        if "detail" in task and task["detail"] == "Not found.":
            self.log.error("Task not found: %s" % args.task_id)
            sys.exit(1)

        task["report"] = "%s/playbooks/%s.html" % (args.server, task["playbook"]["id"])
        columns = (
            "id",
            "report",
            "name",
            "action",
            "status",
            "path",
            "lineno",
            "started",
            "ended",
            "duration",
            "tags",
            "handler",
        )
        return (columns, ([task[column] for column in columns]))


class TaskDelete(Command):
    """ Deletes the specified task and associated resources """

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(TaskDelete, self).get_parser(prog_name)
        parser = global_arguments(parser)
        # fmt: off
        parser.add_argument(
            "task_id",
            metavar="<task-id>",
            help="Task to delete",
        )
        # fmt: on
        return parser

    def take_action(self, args):
        client = get_client(
            client=args.client,
            endpoint=args.server,
            timeout=args.timeout,
            username=args.username,
            password=args.password,
            verify=False if args.insecure else True,
            run_sql_migrations=False,
        )

        # TODO: Improve client to be better at handling exceptions
        client.delete("/api/v1/tasks/%s" % args.task_id)


class TaskMetrics(Lister):
    """ Provides metrics about actions in tasks """

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(TaskMetrics, self).get_parser(prog_name)
        parser = global_arguments(parser)
        # fmt: off
        parser.add_argument(
            "--aggregate",
            choices=["action", "name", "path"],
            default="action",
            help=("Aggregate tasks by action, name or path. Defaults to action."),
        )
        # Task search arguments and ordering as per ara.api.filters.TaskFilter
        parser.add_argument(
            "--playbook",
            metavar="<playbook_id>",
            default=None,
            help=("Filter for tasks for a specified playbook id"),
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            default=None,
            help=("Filter for tasks matching a specific status ('completed', 'expired', 'running' or 'unknown')")
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=("Filter for tasks matching the provided name (full or partial)"),
        )
        parser.add_argument(
            "--path",
            metavar="<path>",
            default=None,
            help=("Filter for tasks matching the provided path (full or partial)"),
        )
        parser.add_argument(
            "--action",
            metavar="<action>",
            default=None,
            help=("Filter for tasks matching a specific action/ansible module (ex: 'debug', 'package', 'set_fact')"),
        )
        parser.add_argument(
            "--long",
            action="store_true",
            default=False,
            help=("Don't truncate paths and include additional status fields: completed, running, expired, unknown")
        )
        parser.add_argument(
            "--order",
            metavar="<order>",
            default="-started",
            help=(
                "Orders tasks by a field ('id', 'created', 'updated', 'started', 'ended', 'duration')\n"
                "Defaults to '-started' descending so the most recent task is at the top.\n"
                "The order can be reversed by omitting the '-': ara task metrics --order=started\n"
                "This influences the API request, not the ordering of the metrics."
            ),
        )
        parser.add_argument(
            "--limit",
            metavar="<limit>",
            default=os.environ.get("ARA_CLI_LIMIT", 1000),
            help=("Return metrics for the first <limit> determined by the ordering. Defaults to ARA_CLI_LIMIT or 1000.")
        )
        # fmt: on
        return parser

    def take_action(self, args):
        client = get_client(
            client=args.client,
            endpoint=args.server,
            timeout=args.timeout,
            username=args.username,
            password=args.password,
            verify=False if args.insecure else True,
            run_sql_migrations=False,
        )
        query = {}
        if args.playbook is not None:
            query["playbook"] = args.playbook

        if args.status is not None:
            query["status"] = args.status

        if args.name is not None:
            query["name"] = args.name

        if args.path is not None:
            query["path"] = args.path

        if args.action is not None:
            query["action"] = args.action

        query["order"] = args.order
        query["limit"] = args.limit

        tasks = client.get("/api/v1/tasks", **query)

        # TODO: This could probably be made more efficient without needing to iterate a second time
        # Group tasks by aggregate
        aggregate = {}
        for task in tasks["results"]:
            item = task[args.aggregate]
            if item not in aggregate:
                aggregate[item] = []
            aggregate[item].append(task)

        data = {}
        for item, tasks in aggregate.items():
            data[item] = {
                "count": len(tasks),
                "results": 0,
                "expired": 0,
                "running": 0,
                "completed": 0,
                "unknown": 0,
                "duration_total": "00:00:00.000000",
            }

            if args.aggregate == "path" and not args.long:
                data[item]["aggregate"] = cli_utils.truncatepath(item, 50)
            else:
                data[item]["aggregate"] = item

            for task in tasks:
                for status in ["running", "completed", "expired", "unknown"]:
                    if task["status"] == status:
                        data[item][status] += 1

                data[item]["results"] += task["items"]["results"]

                if task["duration"] is not None:
                    data[item]["duration_total"] = cli_utils.sum_timedelta(
                        task["duration"], data[item]["duration_total"]
                    )

            data[item]["duration_avg"] = cli_utils.avg_timedelta(data[item]["duration_total"], data[item]["count"])

        # fmt: off
        if args.long:
            columns = (
                "aggregate",
                "count",
                "results",
                "duration_total",
                "duration_avg",
                "completed",
                "running",
                "expired",
                "unknown",
            )
        else:
            columns = (
                "aggregate",
                "count",
                "results",
                "duration_total",
                "duration_avg",
            )

        return (
            columns, (
                [data[action][column] for column in columns]
                for action in sorted(data.keys())
            )
        )
        # fmt: on
